#! /usr/bin/env python3

import re
import sys
import argparse
from Bio.Seq import Seq
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.biotools import fa2dict, gff2list, gff3Comps, dict2fa
from mycotools.lib.kontools import formatPath, sysStart, eprint

def sortGene(sorting_group):

    out_group = []
    for entryType in ['gene', 'mrna', 'trna', 'rrna', 'exon', 'cds']:
        todel = []
        for i, entry in enumerate(sorting_group):
            if entry['type'].lower() == entryType:
                out_group.append(entry)
                todel.append(i)
        for i in reversed(todel):
            sorting_group.pop(i)

    for entry in sorting_group:
        out_group.append(entry)

    return out_group 

def sortContig(contigData):

    coordinates = {}
    for gene in contigData:
        t_coords = []
        for entry in contigData[gene]:
            t_coords.extend([int(entry['start']), int(entry['end'])])
        coordinates[gene] = min(t_coords)
    coordinates = {
        k: v for k, v in sorted(coordinates.items(), key = lambda item: item[1])
        }
    outContig = []
    for gene in coordinates:
        outContig.extend(contigData[gene])

    return outContig



def sortGFF(unsorted_gff, idComp):

    sorting_groups, oldGene = {}, None
    for i, entry in enumerate(unsorted_gff):
        seqid = entry['seqid']
        if seqid not in sorting_groups:
            sorting_groups[seqid] = {}
        gene = idComp.search(entry['attributes'])[1]
        gene = re.sub(r'(.*?_\d+).*', r'\1', gene)
        if gene not in sorting_groups[seqid]:
            sorting_groups[seqid][gene] = []
        sorting_groups[seqid][gene].append(entry)

    sortedGff = []
    for seqid in sorting_groups:
        contigData = {}
        for gene in sorting_groups[seqid]:
            contigData[gene] = sortGene(sorting_groups[seqid][gene])
        sortedGff.extend(sortContig(contigData))

    return sortedGff
 
def sortMain(gff):

    id_comp = re.compile(gff3Comps()['id'])
#    crude_sort = sorted( gff, key = lambda x: \
 #       int( re.search(r'ID=([^;]+)', x['attributes'])[1] ))
    gff = sortGFF(gff, id_comp)

    return gff

def grabCDS( gff_dicts ):
    """Grab CDSs that are associated with genes. gff_dicts is a
    mycotools.lib.biotools gff2list() list"""

    mrnas = []
    for entry in gff_dicts:
        if entry['type'] == 'mRNA':
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            mrnas.extend(alias.split('|')) # account for posttranslational mods

    mrna_set, out_cds = set(mrnas), []
    for entry in gff_dicts:
        if entry['type'] == 'cds':
            alias = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
            if alias in mrna_set:
                out_cds.append(entry)

    return out_cds

def order_neg_dict( neg_dict ):

    for seqid in neg_dict:
        for gene in neg_dict[seqid]:
            for cds in neg_dict[seqid][gene]:
                cds.sort( reverse = True )
            neg_dict[seqid][gene] = sorted( 
                neg_dict[seqid][gene], key = lambda x: x[0], reverse = True
                )
    
    return neg_dict

def grabCoords( cdss ):
    
    pos_dict, neg_dict = {}, {}
    for entry in cdss:
        gene = re.search( gff3Comps()['Alias'], entry['attributes'] )[1]
        seqid = entry['seqid']
        if entry['strand'].rstrip() == '+':
            if seqid not in pos_dict:
                pos_dict[seqid], pos_dict[seqid][gene] = {}, []
            elif gene not in pos_dict[seqid]:
                pos_dict[seqid][gene] = []
            pos_dict[seqid][gene].append( [
                int(entry['start']) - 1, int(entry['end'])
                ] )
        elif entry['strand'].rstrip() == '-':
            negseq = seqid
            if seqid not in neg_dict:
                neg_dict[seqid], neg_dict[seqid][gene] = {}, []
            elif gene not in neg_dict[seqid]:
                neg_dict[seqid][gene] = []
            neg_dict[seqid][gene].append( [
                int(entry['start']), int(entry['end'])
                ] )

    return pos_dict, order_neg_dict(neg_dict)

def translatePos( contig_seq, postig_dict ):

    genes_fa_dict = {}
    for gene in postig_dict:
        nt_seq = ''.join([
            contig_seq[x[0]:x[1]] for x in postig_dict[gene]
            ])
        genes_fa_dict[gene] = {
            'sequence': str(Seq(nt_seq).translate()), 'description': ''
            }

    return genes_fa_dict

def translateNeg( rev_seq, negtig_dict ):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        nt_seq = ''.join([
            rev_seq[seqlen-x[0]:seqlen-x[1]+1] for x in negtig_dict[gene]
            ])
        genes_fa_dict[gene] = {
            'sequence': str(Seq(nt_seq).translate()), 'description': ''
            }

    return genes_fa_dict 

def ntPos( contig_seq, postig_dict ):

    genes_fa_dict = {}
    for gene in postig_dict:
        genes_fa_dict[gene] = {
            'sequence': ''.join([
                contig_seq[x[0]:x[1]] for x in postig_dict[gene]
                ]),
            'description': ''
            }

    return genes_fa_dict


def posPlusMinusCode(genePostig_dict, plusminus, contig_seq, plus = True, minus = True):
    geneDict = {'sequence': '', 'description': ''}
    minimumList = []
    for x in genePostig_dict:
        minimumList.extend(x)
        geneDict['sequence'] += contig_seq[x[0]:x[1]]
    minimum = min(minimumList)
    maximum = max(minimumList)
    if minus:
        if minimum:
            if minimum - plusminus > 0:
                geneDict['sequence'] = contig_seq[minimum-plusminus:minimum] + \
                    geneDict['sequence']
                geneDict['description'] = '-' + str(plusminus)
            else:
                geneDict['sequence'] = contig_seq[0:minimum] + \
                    geneDict['sequence']
                geneDict['description'] = '-' + str(minimum)
    if plus:
        geneDict['sequence'] += contig_seq[maximum+1:maximum+plusminus+1]
        if len(contig_seq) < maximum + plusminus:
            geneDict['description'] += '+' + str(len(contig_seq) - maximum)
        else:
            geneDict['description'] += '+' + str(plusminus)
    return geneDict


def posPlusMinus(genePostig_dict, plusminus, contig_seq, plus = True, minus = True):
    minimumList = []
    for x in genePostig_dict:
        minimumList.extend(x)
    minimum = min(minimumList)
    maximum = max(minimumList)

    geneDict = {'sequence': '', 'description': ''}
    if minus:
        if minimum - plusminus > 0:
            geneDict['sequence'] = contig_seq[minimum-plusminus:maximum+1]
            if plusminus:
                geneDict['description'] = '-' + str(plusminus)
        else:
            geneDict['sequence'] = contig_seq[:maximum+1]
            if plusminus:
                geneDict['description'] = '-' + str(minimum)
    if plus:
        geneDict['sequence'] += contig_seq[maximum+1:maximum+1+plusminus]
        if plusminus:
            if maximum + plusminus > len(contig_seq):
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(len(contig_seq) - maximum)
            else:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(plusminus)
    return geneDict

def ntPosNoncode(contig_seq, postig_dict, plusminus = 0):

    genes_fa_dict = {}
    for i, gene in enumerate(list(postig_dict.keys())):
        genes_fa_dict[gene] = {'sequence': '', 'description': ''}
        if plusminus:
            genes_fa_dict[gene] = posPlusMinus(postig_dict[gene], plusminus, contig_seq)
        else:
            minimumList = []
            for x in postig_dict[gene]:
                minimumList.extend(x)
            minimum = min(minimumList)
            maximum = max(minimumList)
    
            genes_fa_dict[gene]['sequence'] = contig_seq[minimum:maximum]
   
    return genes_fa_dict

def negPlusMinusCode(geneNegtig_dict, seqlen, plusminus, rev_seq, plus = True, minus = True):
    minimumList, geneDict = [], {'sequence': '', 'description': ''}
    for x in geneNegtig_dict:
        minimumList.extend(x)
        geneDict['sequence'] += rev_seq[seqlen-x[0]:seqlen-x[1]+1]
    minimum = min(minimumList)
    maximum = max(minimumList)
    if minus:
        if maximum + plusminus < seqlen:
            geneDict['sequence'] = \
                rev_seq[seqlen-maximum-plusminus:seqlen-maximum] + geneDict['sequence']
            geneDict['description'] = '-' + str(plusminus) + ';+'
        else:
            geneDict['sequence'] = \
                rev_seq[:seqlen-maximum] + geneDict['sequence']
            geneDict['description'] = '-' + str(seqlen-maximum) + ';+'
    if plus:
        if minimum - plusminus > 0:
            geneDict['sequence'] += rev_seq[seqlen-minimum+1:seqlen-minimum+plusminus+1]
            geneDict['description'] += str(plusminus)
        else:
            geneDict['sequence'] += rev_seq[seqlen-minimum+1:]
            geneDict['description'] += str(seqlen-minimum)   
    return geneDict

def negPlusMinus(neg_gene, seqlen, plusminus, rev_seq, plus = True, minus = True):
    minimumList = []
    for x in neg_gene:
        minimumList.extend(x)
    minimum = min(minimumList)
    maximum = max(minimumList)

    geneDict = {'sequence': '', 'description': ''}
    if minus:
        if maximum + plusminus < seqlen:
            geneDict['sequence'] = \
                rev_seq[seqlen-maximum-plusminus:seqlen-maximum]
            if plusminus:
                geneDict['description'] = '-' + str(plusminus)
        else:
            geneDict['sequence'] = \
                rev_seq[:seqlen-maximum]
            if plusminus:
                geneDict['description'] = '-' + str(seqlen-maximum)
    if plus:
        if minimum - plusminus > 0:
            geneDict['sequence'] += rev_seq[seqlen-maximum:seqlen-minimum+plusminus+1]
            if plusminus:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(plusminus)
        else:
            geneDict['sequence'] += rev_seq[seqlen-maximum+1:]
            if plusminus:
                if geneDict['description']:
                    geneDict['description'] += ';'
                geneDict['description'] += '+' + str(seqlen-minimum)   
    return geneDict


def ntNeg( rev_seq, negtig_dict, plusminus = 0 ):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        genes_fa_dict[gene] = {
            'sequence': ''.join([
                rev_seq[seqlen-x[0]:seqlen-x[1]+1] for x in negtig_dict[gene]
                ]),
            'description': ''
            }

    return genes_fa_dict 

def ntNegNoncode(rev_seq, negtig_dict, plusminus = 0):

    genes_fa_dict = {}
    seqlen = len(rev_seq.rstrip())
    for gene in negtig_dict:
        if plusminus:
            genes_fa_dict[gene] = negPlusMinus(negtig_dict[gene], plusminus, seqlen, rev_seq)
        else:
            minimumList = []
            for x in negtig_dict[gene]:
                minimumList.extend(x)
            minimum = min(minimumList)
            maximum = max(minimumList)

            genes_fa_dict[gene] = {
                'sequence': rev_seq[seqlen - maximum:seqlen - minimum + 1],
                'description': ''
                }

    return genes_fa_dict


def ntmain( gff_dicts, assem_dict, coding = True, flanks = True, fullRegion = False, plusminus = 0 ):
    
    if fullRegion:
        flanks, coding = True, False
    contig_dict, contig_info, genes_fa_dict, geneOrder = {}, {}, {}, {}
    cdss = grabCDS( sortMain(gff_dicts) )
    for cds in cdss:
        seqid = cds['seqid']
        if seqid not in contig_dict:
            contig_dict[seqid] = []
            geneOrder[seqid] = []
        contig_dict[seqid].append(cds)
        gene = re.search(gff3Comps()['Alias'], cds['attributes'])[1]
        if gene not in set(geneOrder[seqid]):
            geneOrder[seqid].append(gene)
    for seqid in contig_dict:
        contig_info[seqid] = [(
            re.search(gff3Comps()['Alias'], contig_dict[seqid][0]['attributes'])[1],
            contig_dict[seqid][0]['strand']
            ),(
            re.search(gff3Comps()['Alias'], contig_dict[seqid][-1]['attributes'])[1],
            contig_dict[seqid][-1]['strand']
            )]

    pos_dict, neg_dict = grabCoords( cdss )

    contig_fa_dict, startFlanks, endFlanks = {seqid: {} for seqid in contig_info}, {}, {}
    if flanks:
        if coding:
            for seqid in contig_info:
                startGene, startStrand = contig_info[seqid][0][0], contig_info[seqid][0][1]
                endGene, endStrand = contig_info[seqid][1][0], contig_info[seqid][1][1]
                if startGene == endGene:
                    if startStrand == '+':
                        startFlanks[seqid] = posPlusMinusCode(
                            pos_dict[seqid][startGene],
                            plusminus,
                            assem_dict[seqid]['sequence'],
                            )
                        del pos_dict[seqid][startGene]
                    else:
                        rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                        startFlanks[seqid] = negPlusMinusCode(
                            neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp
                            )
                        del neg_dict[seqid][startGene]
                    continue
                if startStrand == '+':
                    startFlanks[seqid] = posPlusMinusCode(
                        pos_dict[seqid][startGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        plus = False
                        )
                    del pos_dict[seqid][startGene]
                else:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    startFlanks[seqid] = negPlusMinusCode(
                        neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp, plus = False
                        )
                    del neg_dict[seqid][startGene]
                if endStrand == '+':
                    endFlanks[seqid] = posPlusMinusCode(
                        pos_dict[seqid][endGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        minus = False
                        )
                    del pos_dict[seqid][endGene]
                else:
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    endFlanks[seqid] = negPlusMinusCode(
                        neg_dict[seqid][endGene], len(rev_comp), plusminus, rev_comp, minus = False
                        )
                    del neg_dict[seqid][endGene]
        else:
            for seqid in contig_info:
                startGene, startStrand = contig_info[seqid][0][0], contig_info[seqid][0][1],
                endGene, endStrand = contig_info[seqid][1][0], contig_info[seqid][1][1]
                      
                if startGene == endGene:
                    if startStrand == '+':
                        startFlanks[seqid] = posPlusMinus(
                            pos_dict[seqid][startGene],
                            plusminus,
                            assem_dict[seqid]['sequence'],
                            )
                        del pos_dict[seqid][startGene]
                    else:
                        rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                        startFlanks[seqid] = negPlusMinus(
                            neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp
                            )
                        del neg_dict[seqid][startGene]
                    continue

                elif fullRegion:
                    region = []
                    if startStrand == '+':
                        region.append(min(pos_dict[seqid][startGene]))
                    else:
                        region.append(min(neg_dict[seqid][startGene]))
                    if endStrand == '+':
                        region.append(max(pos_dict[seqid][endGene]))
                    else:
                        region.append(max(neg_dict[seqid][endGene]))
                    name = startGene + '-' + endGene
                    genes_fa_dict[name + '_sense'] = posPlusMinus(
                        region, plusminus, assem_dict[seqid]['sequence']
                        )
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    genes_fa_dict[name + '_antisense'] = negPlusMinus(
                        region, len(rev_comp), plusminus, rev_comp
                        )

                
                if startStrand == '+':
                    startFlanks[seqid] = posPlusMinus(
                        pos_dict[seqid][startGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        plus = False
                        )
                    del pos_dict[seqid][startGene]
                else:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    startFlanks[seqid] = negPlusMinus(
                        neg_dict[seqid][startGene], len(rev_comp), plusminus, rev_comp, plus = False
                        )
                    del neg_dict[seqid][startGene]
                if endStrand == '+':
                    endFlanks[seqid] = posPlusMinus(
                        pos_dict[seqid][endGene],
                        plusminus,
                        assem_dict[seqid]['sequence'],
                        minus = False
                        )
                    del pos_dict[seqid][endGene]
                else:
                    rev_comp = str(Seq(assem_dict[seqid]['sequence']).reverse_complement())
                    endFlanks[seqid] = negPlusMinus(
                        neg_dict[seqid][endGene], len(rev_comp), plusminus, rev_comp, minus = False
                        )
                    del neg_dict[seqid][endGene]

        if coding:
            for contig in contig_fa_dict:
                if contig in pos_dict:
                    contig_fa_dict[contig] = { 
                        **ntPos(assem_dict[contig]['sequence'], pos_dict[contig]), 
                        **contig_fa_dict[contig]
                        }
                if contig in neg_dict:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    contig_fa_dict[contig] = {
                        **ntNeg(rev_comp, neg_dict[contig]),
                        **contig_fa_dict[contig]
                        }
        elif not fullRegion:
            for contig in contig_fa_dict:
                if contig in pos_dict:
                    contig_fa_dict[contig] = { 
                        **ntPosNoncode(assem_dict[contig]['sequence'], pos_dict[contig]), 
                        **contig_fa_dict[contig]
                        }
                if contig in neg_dict:
                    rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
                    contig_fa_dict[contig] = {
                        **ntNegNoncode(rev_comp, neg_dict[contig]),
                        **contig_fa_dict[contig]
                        }
    if not fullRegion:
        for contig in geneOrder:
            if contig in startFlanks:
                genes_fa_dict[geneOrder[contig][0]] = startFlanks[contig]
            for gene in geneOrder[contig][1:-1]:
                genes_fa_dict[gene] = contig_fa_dict[contig][gene]
            if contig in endFlanks:
                genes_fa_dict[geneOrder[contig][-1]] = endFlanks[contig]
    
    return genes_fa_dict

def aamain( gff_dicts, assem_dict ):
    
    cdss = grabCDS( gff_dicts )
    pos_dict, neg_dict = grabCoords( cdss )

    genes_fa_dict = {}
    for contig in pos_dict:
        genes_fa_dict = { 
            **translatePos(assem_dict[contig]['sequence'], pos_dict[contig]), 
            **genes_fa_dict 
            }
    for contig in neg_dict:
        rev_comp = str(Seq(assem_dict[contig]['sequence']).reverse_complement())
        genes_fa_dict = {
            **translateNeg(rev_comp, neg_dict[contig]),
            **genes_fa_dict
            }

    return genes_fa_dict

            
if __name__ == '__main__':

    parser = argparse.ArgumentParser(
       description = 'Inputs MycoDB compatible gff3, assembly (optional), ' + \
           'and outputs nucleotides/proteins. Use an abstracted gene gff ' + \
           '(acc2gff.py) to only output a smaller set of gene(s).'
       )
    parser.add_argument('-g', '--gff', required = True)
    parser.add_argument('-n', '--nucleotide', action = 'store_true')
    parser.add_argument('-p', '--protein', action = 'store_true')
    parser.add_argument('-a', '--assembly')
    parser.add_argument('-i', '--intergenic', action = 'store_true', help = '-n only')
    parser.add_argument('-nc', '--noncoding', action = 'store_true', help = '-n only')
    parser.add_argument('-pm', '--plusminus', default = 0, type = int, help = '-n only')
    parser.add_argument('-af', '--all_flanks', action = 'store_true', help = '-n and -nc only')
    args = parser.parse_args()

    input_gff = gff2list(formatPath(args.gff))
    if args.assembly:
        assembly_dicts = {'input': fa2dict( formatPath(args.assembly) )}
        gff_dicts = {'input': input_gff}
    else:
        db = mtdb(masterDB()).set_index('ome')
        gff_dicts, assembly_dicts = {}, {}
        try:
            for line in input_gff:
                gene = re.search(gff3Comps()['Alias'], line['attributes'])[1]
                ome = re.search( r'(.*?)_', gene )[1]
                if ome not in gff_dicts:
                    gff_dicts[ome] = []
                    assembly_dicts[ome] = fa2dict(db[ome]['fna'])
                gff_dicts[ome].append(line)
        except IndexError:
            eprint('\nERROR: ' + args.gff + ' is incompatible with MycotoolsDB', flush = True)
            sys.exit(1)
        
    if args.protein:
        for ome in gff_dicts:
            print(dict2fa(aamain(gff_dicts[ome], assembly_dicts[ome])), flush = True)
    else:
        for ome in assembly_dicts:
            print(dict2fa(
                ntmain(gff_dicts[ome], assembly_dicts[ome], not args.noncoding, not args.all_flanks, args.intergenic, args.plusminus)
                ), flush = True)

    sys.exit(0)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import multiprocessing as mp
from mycotools.lib.kontools import eprint, formatPath, file2list
from mycotools.lib.dbtools import masterDB, mtdb
from mycotools.lib.biotools import gff2list, fa2dict, dict2fa, list2gff, gff3Comps
from mycotools.acc2gff import grabGffAcc


def grabFiles( ome ):
    # grab MycotoolsDB files using ome info
    gff = formatPath( '$MYCOGFF3/' + ome + '.gff3' )
    if not os.path.isfile(gff):
        eprint( 'ERROR: ' + ome + ' gff not detected' , flush = True)
    prot = formatPath( '$MYCOFAA/' + ome + '.aa.fa' )
    return gff, prot


def prepGffOutput( hit_list, gff_path, cpu = 1):

    gff_list = gff2list( gff_path )
    mp_cmds = []
    for hit in hit_list:
        mp_cmds.append( [gff_list, hit] )
    with mp.get_context('spawn').Pool( processes = cpu ) as pool:
        gff_list_strs = pool.starmap(
            grabGffAcc, mp_cmds
        )
    gff_strs = [ list2gff(x) for x in gff_list_strs ]
    gff_str = '##gff-version 3\n'
    for x in gff_strs:
        for line in x.split('\n'):
            if not line.startswith('#'):
                gff_str += line + '\n'

    return gff_str


def prepFaaOutput( hit_list, proteome_path ):

    proteome_dict = fa2dict( proteome_path )
    clus_fa = {}
    for hit in hit_list:
        clus_fa[hit] = proteome_dict[hit]

    return dict2fa( clus_fa )


def compileCDS( gff_list ):

    gene = False
    for entry in gff_list:
        if entry['type'] == 'CDS':
            break
        else:
            entry = None


    # need to make this predict in a better way
    prot_comps = [ 
    r'Alias=([^;]*)', r';Name=(.*?);', r'alias ([^;]*)', r'protein_id\=(.*?;|.*?$)', 
    r'proteinId (\d+)', r'gene_id "(.*?)"', r'Parent=(.*?)$'
    ]
    for tag in prot_comps:
        if re.search( tag, entry['attributes'] ):
            prot_comp = re.compile( tag )
            if tag == prot_comps[-1]:
                gene = True
            break

    if gene:
        mRNA_dict = {}
        for entry in [ x for x in gff_list if x['type'] == 'mRNA' ]:
            gene = re.search( r'Alias=([^;]*)', entry['attributes'] )[1]
            mRNA = re.search( r'ID=([^;]*)', entry['attributes'] )[1]
            mRNA_dict[mRNA] = gene


    cds_dict = {}
    for entry in gff_list:
        if entry['type'] == 'CDS':
            if entry['seqid'] not in cds_dict:
                cds_dict[ entry['seqid'] ] = {}
            prot = prot_comp.search( entry['attributes'] )[1]
            if gene:
                prot = mRNA_dict[ prot ]
            if prot not in cds_dict[ entry['seqid'] ]:
                cds_dict[ entry['seqid'] ][ prot ] = []
            cds_dict[ entry['seqid'] ][ prot ].extend( [int( entry['start']), int( entry['end'] )] )
            
    for seqid in cds_dict:
        for prot in cds_dict[ seqid ]:
            cds_dict[ seqid ][ prot ].sort()
        cds_dict[seqid] = sorted( cds_dict[seqid].keys(), key = lambda k: cds_dict[seqid][k][0] )

    return cds_dict


def compileCDS_mycotools(gff_list):
    '''
    Inputs the gff_list and organism ome code. Compiles CDS entries for loci parsing and outputs:
    cds_dict = {contig: {protein: [[CDSstart, CDSstop]] } }
    Sorts the cds_dict for each protein from smallest to largest
    '''

    cds_dict, fail = {}, False
    for entry in gff_list:
        if entry['type'] == 'CDS':
            if entry['seqid'] not in cds_dict:
                cds_dict[entry['seqid']] = {}
            prot_prep_i0 = entry['attributes'].index(';Alias=')
            try:
                prot_prep_i1 = entry['attributes'][prot_prep_i0+1:].index(';') + prot_prep_i0+1
            except ValueError:
                prot_prep_i1 = len(entry['attributes'])
            prot = entry['attributes'][prot_prep_i0:prot_prep_i1].replace(';Alias=','')
            if prot not in cds_dict[entry['seqid']] and prot:
                cds_dict[entry['seqid']][prot] = []
            elif not prot:
                print(entry['attributes'], prot_prep_i0, prot_prep_i1)
                if not fail:
                    print('\tWARNING: proteins in gff with no Alias', flush = True)
                fail = True
                continue
            cds_dict[entry['seqid']][prot].extend([int(entry['start']), int(entry['end'])])

    for seqid in cds_dict:
        for prot in cds_dict[seqid]:
            cds_dict[seqid][prot].sort()
        cds_dict[seqid] = list(sorted(cds_dict[seqid].keys(), key = lambda k: cds_dict[seqid][k][0]))

    return cds_dict


def prepOutput( cds_dict, accession, plusminus ):

    out_index = []
    for seqid in cds_dict:
        if accession in set(cds_dict[ seqid ]):
            index = cds_dict[ seqid ].index( accession )
            if index - plusminus < 0:
                lower = 0
            else:
                lower = int(index - plusminus)
            upper = int( index + plusminus ) + 1
            for index in cds_dict[ seqid ][ lower:upper ]:
                out_index.append( index )
            break    

    return out_index


def main( gff_list, accessions, plusminus = 10, mycotools = False, geneGff = False ):

    out_indices = {}
    if mycotools:
        cds_dict = compileCDS_mycotools(gff_list)
    else:
        cds_dict = compileCDS( gff_list )
    for accession in accessions:
        out_indices[accession] = prepOutput( cds_dict, accession, plusminus )

    out_indices = {k: v for k, v in out_indices.items() if v}
    if geneGff:
        geneGffs_prep = {acc: {} for acc in out_indices}
        gene_sets = {acc: set(genes) for acc, genes in out_indices.items()}
        for entry in gff_list:
            if 'RNA' in entry['type']:
                try:
                    gene = re.search(gff3Comps()['Alias'], entry['attributes'])[1]
                    for acc, genes in gene_sets.items():
                        if gene in genes:
                            geneGffs_prep[acc][gene] = entry
#                            break # if one gene is in multiple loci it needs to show up
                except TypeError: # no alias
                    pass

        geneGffs, todel = {}, []
        for acc in out_indices:
            geneGffs[acc] = []
            for gene in out_indices[acc]:
                try:
                    geneGffs[acc].append(geneGffs_prep[acc][gene])
                except KeyError: # gene without RNA
                    raise KeyError('gene without RNA')
#                    todel.append((acc, gene))
        for acc, gene in todel:
            del out_indices[acc][gene]
        return out_indices, geneGffs
    return out_indices


def mycotools_main(db, accessions, plusminus = 10, cpus = 1):

    acc_dict = {}
    for acc in accessions:
        ome = acc[:acc.find('_')]
        if ome not in acc_dict:
            acc_dict[ome] = []
        acc_dict[ome].append(acc)

    db = db.set_index('ome')
    cmds = [
        [gff2list(db[ome]['gff3']), accs, args.plusminus, True] for ome, accs in acc_dict.items()
        ]
    with mp.get_context('spawn').Pool(processes = cpu) as pool:
        acc_res = pool.starmap(main, cmds)

    out_indices = {}
    for res in acc_res:
        out_indices = {**out_indices, **res}

    return out_indices


if __name__ == '__main__':

    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( description = 'Extracts loci from accession(s)' )
    parser.add_argument( '-i', '--input', help = 'File of accessions' )
    parser.add_argument( '-a', '--accession' )
    parser.add_argument( '-p', '--plusminus', default = 10, type = int,
        help = 'Genes +/- from accession. DEFAULT: 10' )
    parser.add_argument( '-o', '--output', action = 'store_true',
        help = 'Output locus fasta(s) and gff(s)' )
    parser.add_argument( '-g', '--gff', help = 'Input GFF file' )
    parser.add_argument( '-f', '--faa', help = 'Input protein fasta file' )
    parser.add_argument( '-s', '--sep', help = 'Separator for input file.', default = '\n')
    parser.add_argument( '--cpu', type = int, default = 1, help = 'CPUs, DEFAULT: 1' )
    args = parser.parse_args()

    if args.cpu < mp.cpu_count():
        cpu = args.cpu
    else:
        cpu = mp.cpu_count()
    args.sep = args.sep.replace("'",'').replace('"','')
 
    if args.input:
        accessions = file2list( formatPath(args.input), sep = args.sep )
    elif args.accession:
        accessions = [ args.accession ]
    else:
        print('\nERROR: requires input or accession', flush = True)
        sys.exit( 1 )

    db = None

    out_indices = {}
    if args.gff: 
        gff = gff2list( formatPath(args.gff) )
#        if not args.faa:
 #           print('\nERROR: no proteome', flush = True)
  #          sys.exit( 3 )
        out_indices = main( gff, accession, args.plusminus )
    else:
        db = mtdb( formatPath( masterDB() ) ).set_index('ome')
        out_indices = mycotools_main(db, accessions, plusminus = 10, cpus = args.cpu)

#        for i in acc_res:
 #           out_indices[i[1]] = i[0] 

    if args.output:
        if not db:
            db = mtdb(formatPath(masterDB())).set_index('ome')
        for accession in out_indices:
            if args.gff:
                gff = formatPath(args.gff)
                if args.faa:
                    prot = formatPath(args.faa)
                else:
                    prot = None
            else:
                ome = accessions[:accession.find('_')]
                gff, prot = db[ome]['gff3'], db[ome]['faa']
#            for hit in out_indices[accession]:
            if len( out_indices[accession] ) > 0:
                gff_str = prepGffOutput( 
                    out_indices[accession], gff, cpu = args.cpu
                )
                with open( accession + '.locus.gff3', 'w' ) as out:
                    out.write( gff_str )
                if prot:
                    prot_str = prepFaaOutput(
                        out_indices[accession], prot
                        )
                    with open( accession + '.locus.aa.fa', 'w' ) as out:
                        out.write( prot_str )
    else:
        for accession in out_indices:
#            print('\n' + accession + ' locus +/- ' + str(args.plusminus), flush = True)
            for hit in out_indices[accession]:
#                for index in out_indices[accession]:
                print( hit, flush = True)
            print(flush = True)
    sys.exit( 0 )



#! /usr/bin/env python3

import os
import re
import sys
from mycotools.lib.biotools import gff2list
from mycotools.lib.kontools import formatPath, eprint


def compileExon( gff_path, output, ome = None ):
    '''
    Inputs: `gff_path` or mycotoolsDB file
    Outputs: summary annotation statistics
    Import the gff, find the first exon, and test which protein regular
    expression pattern to use. For each line in the gff, if it is an exon grab
    the protein and include it in `exon_dict`. Extend the start and stop
    coordinates for that exon. For each protein in the `exon_dict`, sort the
    entry, add the exon length to the total by subtracting the smallest entry
    from the largest for that protein. Append this value to the length list for
    median evaluation. Sort the `len_list` and calculate/output annotation
    statistics.
    '''

    gff = gff2list( gff_path )
    exon_dict = {}

    for index in range(len(gff)):
        if gff[index]['type'].lower() == 'exon':
            break

    protComp = re.compile( r';Parent\=([^;]*)' )
    if not protComp.search( gff[index]['attributes'] ):
        protComp = re.compile( r'gene_id "(.*?)"' )
        if not protComp.search( gff[index]['attributes'] ):
            protComp = re.compile( r'name "(.*?)"\;' )
            if not protComp.search( gff[index]['attributes'] ):
                protComp = re.compile( r'ID=(.*?);' )
    for line in gff:
        if line['type'].lower() == 'exon':
            prot = protComp.search( line['attributes'] )[1]
            if prot not in exon_dict:
                exon_dict[ prot ] = []
            exon_dict[ prot ].extend( [ int( line['start'] ), int( line['end'] ) ] )

    total, len_list = 0, []
    for prot in exon_dict:
        exon_dict[ prot ].sort()
        total += exon_dict[prot][-1] - exon_dict[prot][0]
        len_list.append( exon_dict[prot][-1] - exon_dict[prot][0] )

    len_list.sort()
    try:
        if len( len_list ) % 2 == 0:
            median = len_list[int(len(len_list)/2 - 1)]
        else:
            median = (len_list[round(len(len_list)/2 - 1)] + len_list[round(len(len_list)/2 - 2)])/2
    except IndexError:
        eprint('ERROR: ' + os.path.basename(gff_path) + ' - no exons detected. Skipping', flush = True)
        return None

    if any( line for line in gff if line['type'] == 'intron' ):
        eprint('ERROR: ' + os.path.basename(gff_path) + ' - introns detected. Convert to exons. Skipping', flush = True)
        return None

    if not output:
        print( '{:<25}'.format('GENE LENGTH:') + str(total) , flush = True)
        print( '{:<25}'.format('GENES:') + str(len(exon_dict)), flush = True)
        print( '{:<25}'.format('MEAN GENE LENGTH:') + str(total/len(exon_dict)), flush = True)
        print( '{:<25}'.format('MEDIAN GENE LENGTH:') + str(median), flush = True)
    geneStats = {
        'total_len': total, 'genes': len(exon_dict), 
        'mean_len': total/len(exon_dict), 'median_len': median 
        }

    return ome, geneStats


def main(in_path, log_path = None, cpus = 1):

    if in_path[-4:] not in {'.gtf', '.gff', 'gff3'}:
        from mycotools.lib.dbtools import mtdb
        import multiprocessing as mp
        db = mtdb(in_path).set_index()

        prevOmes = {}
        if log_path and os.path.isfile(log_path):
             with open(log_path, 'r') as raw:
                 for line in raw:
                     if not line.startswith('#'):
                         omeI = line.index('\t')
                         ome = line[:omeI]
                         prevOmes[ome] = line[omeI+1:].rstrip()

        cmds = []
        for ome in db:
            if ome not in prevOmes:
                cmds.append((db[ome]['gff3'], True, ome,))
        with mp.Pool(processes = cpus) as pool:
            res = pool.starmap(compileExon, cmds)

        outPrep = {}
        for result in res:
            if result:
                outPrep[result[0]] = '\t'.join([str(x) for x in result[1].values()])
        out = {
            k: v for k, v in \
            sorted({**outPrep, **prevOmes}.items(), key = lambda x: x[0])
            }


        if not log_path:
#            output_file = os.path.basename(formatPath(sys.argv[1])) + '.annStats.tsv'
            print('#ome\ttotal_length\tgenes\tmean_length\tmedian_length', flush = True)
            for ome in out:
                print(ome + '\t' + out[ome], flush = True)
        else:
            with open(log_path, 'w') as write:
                write.write('#ome\ttotal_length\tgenes\tmean_length\tmedian_length\n')
                for ome in out:
                    write.write(ome + '\t' + out[ome] + '\n')
    else:
        ome, geneStats = compileExon(in_path, log_path)
        if log_path:
            out_str = 'total_length\tgenes\tmean_length\tmedian_length\n' + str(geneStats['total_len']) + \
                '\t' + str(geneStats['genes']) + '\t' \
                + str(geneStats['mean_len']) + '\t' + str(geneStats['median_len'])


if __name__ == '__main__':

    output = False
    usage = '\nUSAGE: `gff`/`gtf`/`gff3` OR mycotoolsDB, optional output file\n'
    if '-h ' in sys.argv or '--help' in sys.argv or '-h' == sys.argv[-1]:
        print(usage, flush = True)
        sys.exit(1)
    elif len( sys.argv ) < 2:
        print(usage, flush = True)
        sys.exit(1)
    elif not os.path.isfile( formatPath(sys.argv[1]) ):
        print(usage, flush = True)
        sys.exit(1)
    elif len( sys.argv ) > 2:
        log_path = formatPath(sys.argv[2])
    else:
        log_path = None

    in_path = formatPath(sys.argv[1])
    main(in_path, log_path, os.cpu_count())

    sys.exit(0)



#! /usr/bin/env python3

import os
import re
import sys
from mycotools.lib.kontools import formatPath, sysStart, eprint
from mycotools.lib.dbtools import masterDB, mtdb

forbidden = [
    '/', '\\', '[', ']', '|', '+', '=', '(', ')',
    '{', '}', ':', ';', '<', '>', '?', '&', '*', '^',
    '%', '@', '#', '$', '~'
    ]

def parseArgs(args):

    genus, species, strain, ome_code, goOn, alternative = True, True, True, \
        True, True, True
    allowable = {
        '_', '-', '!',
        '`', ',', '.',
        '~', "'", '"'
        }

    for arg in args[2:]:
        if len( arg ) > 1:
            if arg[0] in { '"', "'" }:
                arg = arg[1:]
                if arg[-1] in {'"', "'"}:
                    arg = arg[:-1]
        if os.path.isfile( formatPath(arg) ):
            if not goOn:
                eprint( '\nERROR: multiple files' , flush = True)
            db = mtdb( arg )
            goOn = False
            continue
        for let in arg:
            if let == 'g':
                genus = False
            elif let == 's':
                species = False
            elif let == 'v':
                strain = False
            elif let == 'o':
                ome_code = False
            elif let == 'a':
                alternative = False
            elif let in allowable:
                forbidden.append( let )

    if goOn:
        db = mtdb( masterDB() )

    with open( args[1], 'r' ) as raw_input:
        data_input = raw_input.read()

    return db, data_input, genus, species, strain, ome_code, alternative

def main( db, data_input, genus, species, strain, ome_code, alternative ):

    db = db.set_index()
    for ome, row in db.items():
        name = ''
        if genus:
            name += str(row['genus']) + '_'
        if species:
            if row['species']:
                name += str(row['species']) + '_'
            else:
                name += 'sp._'
        if strain:
            if row['strain']:
                name += str(row['strain']) + '_'
        if alternative:
            name += row['assembly_acc'] + '_'
        if ome_code:
            name += ome + '_'
        name = name[:-1]
        for i in forbidden:
            name = name.replace( i, '' )
        data_input = re.sub( ome, name, data_input )

    return data_input.rstrip()


if __name__ == '__main__':

    usage = 'USAGE: ome2name.py <INPUTFILE> | ome2name.py <INPUTFILE>' \
        + ' [MYCODB] asvg*&\nDEFAULTS: master db, see script for default' \
        + ' forbidden characters' + \
        '\nInput file to regex sub omes with their name.\n' + \
        'optional mycotools db, string of forbidden characters\n' + \
        '"o" no ome | "g" no genus | "s" no species | "v" no strain' + \
        ' | "a" no alternative ome'
    args = sysStart( sys.argv, usage, 2, files = [sys.argv[1]] )
    db, data_input, g, sp, st, ome_code, alt = parseArgs(args)
    data_output = main(db, data_input, g, sp, st, ome_code, alt)
    print( data_output , flush = True)
    sys.exit( 0 )



#! /usr/bin/env python3

import argparse, subprocess, os, datetime, sys, re
import multiprocessing as mp, pandas as pd, numpy as np
from mycotools.lib.kontools import eprint, intro, outro, formatPath, multisub, collect_files, findExecs
from mycotools.lib.dbtools import db2df, masterDB
from mycotools.acc2fa import famain as acc2fa


def compileBlastCmd( ome, biofile, env_dir, out_dir, blast_scaf ):
    return blast_scaf + ['-out', out_dir + ome + '.tsv', \
        '-subject', env_dir + biofile]

def compBlastTups( 
    seq_db, blast_type, seq_type, out_dir, 
    biotype, env_dir, query, hsps = None, max_hits = None,
    evalue = None
    ):

    blast_scaf = [
        blast_type, '-query', query,
        '-outfmt', '6'
        ]
    if evalue:
        blast_scaf.extend(['-evalue', str(evalue)])
    if hsps:
        blast_scaf.extend(['-max_hsps', str(hsps)])
    if max_hits:
        blast_scaf.extend([ '-max_target_seqs', str(max_hits) ])

    blast_cmds = []
    for i, row in seq_db.iterrows():
        if not pd.isnull(row[biotype]):
            blast_cmds.append( compileBlastCmd(
                row['internal_ome'], row[biotype], env_dir,
                out_dir, blast_scaf
                ) )

    return blast_cmds


def parseOutput( ome, file_, bitscore = 0, pident = 0 ):

    ome_results = [ome, []]
    if os.path.exists( file_ ):
        with open( file_, 'r' ) as raw:
            raw_data = raw.read()
        data = [x.split('\t') for x in raw_data.split('\n') if x]
        for i in data:
            if int(float(i[-1])) > bitscore and int(1000*float(i[2])) > 100000 * pident:
                ome_results[-1].append(i)

    return ome_results


def compileResults( res_dict ):

    output_res = {}
    for i in res_dict:
        for hit in res_dict[i]:
            query = hit[0]
            subject = hit[1]
            pident = hit[2]
            start = hit[-4]
            end = hit[-3]
            if query not in output_res:
                output_res[query] = {}
            if i not in output_res[query]:
                output_res[query][i] = []

            output_res[query][i].append( [ subject, start, end ] )

    return output_res


def compAcc2fa( db, biotype, env_dir, output_res, subhit = False ):

    acc2fa_cmds, db, subs = {}, db.set_index('internal_ome'), set()
    if subhit or biotype == 'assembly':
        for query in output_res:
            acc2fa_cmds[query] = []
            for i in output_res[query]:
                temp_df, accs = pd.DataFrame(), []
                for hit in output_res[query][i]:
                    accs.append(hit[0] + '[' + hit[1] + '-' + hit[2] + ']')
                temp_df[0] = accs
                if biotype == 'assembly':
                    acc2fa_cmds[query].append(
                        [temp_df, env_dir + db[biotype][i], 0, i]
                        )
                else:
                    acc2fa_cmds[query].append(
                        [temp_df, env_dir + db[biotype][i], 0]
                        )
    else:
        for query in output_res:
            acc2fa_cmds[query] = []
            for i in output_res[query]:
                temp_df, accs = pd.DataFrame(), []
                for hit in output_res[query][i]:
                    subject = hit[0]
                    if subject not in subs:
                        subs.add( subject )
                        accs.append(subject)
                temp_df[0] = accs
                if accs:
                    try:
                        acc2fa_cmds[query].append( [
                            temp_df, env_dir + db[biotype][i], 0
                            ] )
                    except KeyError:
                        eprint('\t' + i + ' not in db')

    return acc2fa_cmds


def prepOutput(out_dir, query):

    out_dir, query = formatPath(out_dir), formatPath( query )
    if not out_dir.endswith('/'):
        out_dir += '/'
    if not os.path.isdir( out_dir ):
        os.mkdir( out_dir )
    report_dir = out_dir + 'reports/'
    if not os.path.isdir( report_dir ):
        os.mkdir( report_dir )

    return report_dir


def ObyOprep(
    db, report_dir, blast, query, 
    max_hits, evalue, out_dir
    ):

    prev, finished, rundb = False, set(), db
    log_str = report_dir + '\n' + blast + '\n' + query + '\n' + \
        str(max_hits) + '\n' + str(evalue)
    log_name = out_dir + '.' + os.path.basename(out_dir[:-1]) + '.log'
    if not os.path.isfile( log_name ):
        with open( log_name, 'w' ) as out:
            out.write( log_str )
    else:
        with open( log_name, 'r' ) as raw:
            log = raw.read()
        if log.rstrip() == log_str.rstrip():
            eprint('\nPicking up previous run', flush = True)
            prev = True
        else:
            eprint('\nInconsistent run parameters in log, rerunning', flush = True)
            count = 0
            while os.path.isdir( report_dir ):
                count += 1
                report_dir = report_dir[:-1] + str(count) + '/'
            os.mkdir( report_dir )
            log_str = report_dir + '\n' + blast + '\n' + query + '\n' + \
                str(max_hits) + '\n' + str(evalue)  
            with open( log_name, 'w' ) as out:
                out.write( log_str )
    if prev:
        reports = collect_files(report_dir, 'tsv')
        finished = {
            os.path.basename(x)[:-4] for x in reports \
            if os.path.getsize(x) > 0
            }
        rundb = db.loc[~db['internal_ome'].isin(finished)]

    return rundb


def ObyOblast(
    rundb, blast, seq_type, report_dir, biotype,
    env_dir, query, hsps, max_hits, evalue, cpus,
    bitscore, pident
    ):

# NEED to add other threshold options
    if len(rundb) > 0:
        print('\nBlasting', flush = True)
        blast_tups = compBlastTups(
            rundb, blast, seq_type, 
            report_dir, biotype, env_dir, query, 
            hsps = hsps, max_hits = max_hits, evalue = evalue
            )

        blast_outs = multisub( blast_tups, processes = cpus )
   
    parse_tups = []
    for i, row in db.iterrows():
        if not pd.isnull(row[biotype]):
            parse_tups.append( [row['internal_ome'],
                report_dir + row['internal_ome'] + '.tsv',
                bitscore, pident ] )

    print('\nParsing reports', flush = True)
    with mp.get_context( 'spawn' ).Pool( processes = cpus ) as pool:
        results = pool.starmap( parseOutput, tuple(parse_tups) )
    results_dict = {x[0]: x[1] for x in results}

    return results_dict


def checkBlastDB():

    db_date = os.path.basename(masterDB())
    blast_db = formatPath('$MYCOFAA/blastdb/' + db_date + '.00.psd')
    if os.path.isfile(blast_db):
        return blast_db[:-7]
    elif os.path.isfile(blast_db[:-7] + '.psd'):
        return blast_db[:-7]


def dbBlast(
    db_path, blast_type, query, 
    evalue, hsps, cpus,
    report_dir
    ):

    out_file = report_dir + os.path.basename(db_path)[:-3] + '.out'
    blast_scaf = [
        blast_type, '-query', query, '-outfmt', '6',
        '-db', db_path, '-num_threads', str(cpus), '-out',
        out_file, '-num_descriptions', str(135904025),
        '-num_alignments', str(135904025)
        ]

    if evalue:
        blast_scaf.extend(['-evalue', str(evalue)])
    if hsps:
        blast_scaf.extend(['-max_hsps', str(hsps)])

    blast_call = subprocess.call(
        blast_scaf#, stdout = subprocess.PIPE,
    #    stderr = subprocess.PIPE
        )

    return blast_call, out_file


def parseDBout( db, file_, bitscore = 0, pident = 0, max_hits = None ):

    ome_results = {}
    with open( file_, 'r' ) as raw:
        for line in raw:
            data = [x for x in line.rstrip().split('\t')]
            ome = re.search(r'(^[^_]*)', data[1])[1]
            if int(float(data[-1])) > bitscore and int(1000*float(data[2])) > 100000 * pident:
                if ome not in ome_results:
                    ome_results[ome] = []
                ome_results[ome].append(data)

    x_omes = set(db['internal_ome'])
    omes_results = {
        x: ome_results[x] for x in ome_results if x in x_omes
        }

    if max_hits:
        out_results = {}
        for ome in ome_results:
            ome_results[ome].sort(key = lambda x: float(x[-1]), reverse = True)
            out_results[ome] = ome_results[ome][:max_hits]
        return out_results
    else:
        for ome in ome_results:
            ome_results[ome].sort(key = lambda x: float(x[-1]), reverse = True)
            return ome_results


def main( 
    db, blast, query, out_dir, hsps = 1, 
    max_hits = None, evalue = 10, bitscore = 0, 
    pident = 0, #coverage = 0, 
    cpus = 1, force = False
    ):

    if blast in {'tblastn', 'blastp'}:
        seq_type = 'prot'
        biotype = 'proteome'
        env_dir = os.environ['MYCOFAA'] + '/'
        blast_db = checkBlastDB()
    elif blast in {'blastx', 'blastn'}:
        seq_type = 'nucl'
        biotype = 'assembly'
        env_dir = os.environ['MYCOFNA'] + '/'
        blast_db = None
    else:
        eprint('\nERROR: invalid blast type: ' + blast, flush = True)

    report_dir = prepOutput(out_dir, query)
    if blast_db and not force:
        print('\nBlasting using MycotoolsDB blastdb')
        blast_exit, blast_output = dbBlast(
            blast_db, blast, query, evalue, 
            hsps, cpus, report_dir
            )
        if blast_exit:
            eprint('\nERROR: blast failed: ' + str(blast_exit))
            sys.exit(10)
        results_dict = parseDBout(
            db, blast_output, bitscore = bitscore, 
            pident = pident, max_hits = max_hits
            )
    else:
        print('\nBlasting on an ome-by-ome basis') 
        rundb = ObyOprep(
            db, report_dir, blast, query, 
            max_hits, evalue, out_dir
            )
        results_dict = ObyOblast(
            rundb, blast, seq_type, report_dir, biotype,
            env_dir, query, hsps, max_hits, evalue, cpus,
            bitscore, pident
            )

    print('\nCompiling fastas', flush = True)
    output_res = compileResults( results_dict )
    acc2fa_cmds = compAcc2fa( db, biotype, env_dir, output_res )
    output_fas = {}
    for query in acc2fa_cmds:
        print('\t' + query, flush = True)
        with mp.get_context('spawn').Pool(processes = cpus) as pool:
            results = pool.starmap( acc2fa, acc2fa_cmds[query] )
        output_fas[query] = '\n'.join(results)

    return output_fas


if __name__ == '__main__':
# NEED abstract covered portion
# NEED max hits after blast compiled
# NEED hsp option
    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( 
        description = 'Blasts a query against a db and compiles output fastas ' + \
            'for each query. --evalue and --maxhit thresholds are applied at ' + \
            'the BLAST step and are therefore incompatible with --previous. Be ' +
            'aware that db2blast.py will preferably blast against a blastdb of ' +
            'Mycotools -omes; however, if the blastdb was not generated from ' +
            'the current masterDB (' + os.path.basename(masterDB()) + ') then ' +
            'db2blast.py will blast on an ome-by-ome basis. In the latter case, ' +
            'e-value is dependent on database size, so it will threshold ' +
            'variably for each ome. If no blastdb exists for the current masterDB ' +
            'in ' + os.environ['MYCOFAA'] + '/blastdb/ then it is recommended to ' +
            'use the bit score threshold as it is not dependent on database size.'
        )
    parser.add_argument( '-b', '--blast', required = True, 
        help = 'Blast type { blastn, blastp, tblastn, blastx }' )
    parser.add_argument( '-q', '--query', required = True, help = 'Query fasta' )
    parser.add_argument( '-e', '--evalue', type = int,
        help = 'Negative e-value order max, e.g. 2 for 10^-2' )
    parser.add_argument( '-s', '--bitscore', default = 0,
        type = int, help = 'Bit score minimum' )
    parser.add_argument( '-i', '--identity', default = 0,
        type = float, help = 'Identity minimum, e.g. 0.6' )
    #parser.add_argument( '-c', '--coverage', type = float, help = 'Query coverage +/-, e.g. 0.5' )
    parser.add_argument( '-m', '--maxhits', type = int, help = 'Max hits for each organism' )
    parser.add_argument( '-d', '--database', default = masterDB(), 
        help = 'DEFAULT: masterdb' )
    parser.add_argument( '-o', '--output', default = os.getcwd() )
    parser.add_argument( '--cpu', default = mp.cpu_count(), type = int, help = 'DEFAULT: all' )
    parser.add_argument( '-f', '--force', action = 'store_true', help = 'Force ome-by-ome blast' )
    args = parser.parse_args()

    db_path = formatPath( args.database )
    if args.cpu > mp.cpu_count():
        cpus = mp.cpu_count()
    else:
        cpus = args.cpu
  
    evalue = None 
    if args.evalue:
        evalue = 10**( -args.evalue)
        
    start_args = {
        'Blast': args.blast, 'query': formatPath(args.query), 
        'database': db_path, 'output': args.output, 'max evalue': evalue,
        'min bitscore': args.bitscore, 'max hits/organism': args.maxhits, #'coverage': args.coverage, 
        'min identity': args.identity, 'cores': cpus, 'force ome-by-ome': args.force
        }

    start_time = intro( 'db2blast', start_args )
    date = start_time.strftime('%Y%m%d')
    findExecs( [args.blast], exit = {args.blast} )
   
    output_dir = formatPath( args.output ) + date + '_db2blast/'

    db = db2df(db_path)

    output_fas = main( 
        db, args.blast, args.query, output_dir, 
        bitscore = args.bitscore, pident = args.identity,
        max_hits = args.maxhits, cpus = cpus, force = args.force 
        )
    if not os.path.isdir( output_dir + 'fastas/' ):
        os.mkdir( output_dir + 'fastas/' )

    for query in output_fas:
        with open( output_dir + 'fastas/' + query + '.blast.fa', 'w' ) as out:
            out.write( output_fas[query] )

    outro( start_time )



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import pandas as pd
from shutil import copy as cp
from mycotools.lib.dbtools import db2df, df2db, masterDB
from mycotools.lib.kontools import formatPath, prep_output


def softMain( filetypes, envtypes, db, output_path, print_link = False ):

    filetypes = {x: filetypes[x] for x in filetypes if filetypes[x]}
    if not print_link:
        for ftype in filetypes:
            if not os.path.isdir(output_path + ftype):
                os.mkdir(output_path + ftype)
        for ftype in filetypes:
            check_db = db[~db[ftype].isnull()]
            check_db.apply(
                lambda x: os.symlink(envtypes[ftype] + '/' + x[ftype],
                output_path + ftype + '/' + x[ftype]), axis = 1
                )
        
    else:
        for i, row in db.iterrows():
            for ftype in filetypes:
                if not pd.isnull( row[ ftype ] ):
                    fpath = formatPath( '$' + envtypes[ftype] + '/' + row[ftype] )
                    print( fpath , flush = True)


def hardMain( filetypes, envtypes, db, output_path ):

    filetypes = {x: filetypes[x] for x in filetypes if filetypes[x]}
    for ftype in filetypes:
        if not os.path.isdir( output_path + ftype ):
            os.mkdir( output_path + ftype )

    for i, row in db.iterrows():
        for ftype in filetypes:
            if not pd.isnull( row[ ftype ] ):
                fpath = formatPath( '$' + envtypes[ftype] + '/' + row[ftype] )
                cp( fpath, output_path + ftype + '/' + os.path.basename( fpath ) ) 

if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = 'Symlinks/copies selected files from database' )
    parser.add_argument( '-d', '--database', default = masterDB(), help = 'DEFAULT: masterDB' )
    parser.add_argument( '-a', '--assembly', action = 'store_true', help = 'Grab assemblies' )
    parser.add_argument( '-p', '--proteome', action = 'store_true', help = 'Grab proteomes' )
    parser.add_argument( '-g', '--gff', action = 'store_true', help = 'Grab gff`s' )
    parser.add_argument( '--print', action = 'store_true', help = 'Print paths, no copy' )
    parser.add_argument( '--hard', action = 'store_true', help = 'Hard copy files' )
    parser.add_argument( '-o', '--output', default = os.getcwd() )
    args = parser.parse_args()

    if not args.assembly and not args.proteome and not args.gff:
        print('\nERROR: no file type (-a, -g, -p) selected', flush = True)
        sys.exit(4)

    db_path = formatPath( args.database )
    output_path = prep_output( formatPath( args.output ), cd = False )
    args_dict = {
        'DATABASE': db_path, 'OUTPUT': output_path,
        'ASSEMBLY': args.assembly, 'PROTEOME': args.proteome,
        'GFF3': args.gff, 'Print links': args.print, 'Hard copy': args.hard
        }

    filetypes = { 
        'assembly': args.assembly, 'proteome': args.proteome,
        'gff3': args.gff
        }
    envtypes = { 
        'assembly': 'MYCOFNA', 'proteome': 'MYCOFAA',
        'gff3': 'MYCOGFF3'
        }

    db = db2df( db_path )
    if args.print or not args.hard:
        softMain(filetypes, envtypes, db, output_path, print_link = args.print)
    else:
        hardMain( filetypes, envtypes, db, output_path )

    sys.exit( 0 )



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import multiprocessing as mp
from mycotools.lib.biotools import gff2list, list2gff
from mycotools.lib.dbtools import mtdb, masterDB
from mycotools.lib.kontools import formatPath

def grabGffAcc( gff_list, acc, term = 'Alias=' ):
    '''grab acc from alias'''

    alias = term + acc
    alias_on = alias + ';'
#    elif ' alias "' in gff_list[0]['attributes']:
 #       alias = 'alias "' + acc + '"'
  #      alias_on = 'alias "' + acc + '"'
    out_list = [ 
        x for x in gff_list \
        if x['attributes'].endswith(alias) or \
            alias_on in x['attributes'] 
        ]
    return out_list


def grabGffAccs(gff_list, acc_list, ome = None):

    aliases, ends = set(), set()
    for acc in acc_list:
        aliases.add('Alias=' + acc)
        ends.add('Alias=' + acc + ';')
    out_list = []
    for i in gff_list:
        if any(i['attributes'].endswith(x) for x in aliases):
            out_list.append(i)
        elif any(x in i['attributes'] for x in ends):
            out_list.append(i)

    if ome:
        return out_list, ome
    else:
        return out_list


def gffMain(gffData, accs):
    accGffs = {}
    if isinstance(gffData, str):
        gff = gff2list(gffData)
    else:
        gff = gffData
    for acc in accs:
        accGffs[acc] = grabGffAcc(gff, acc)
    return accGffs

def dbMain(db, accs, cpus = 1):

    omes = set([x[:x.find('_')] for x in accs])
    db = db.set_index('ome')
    grabAcc_cmds = []
    for ome in list(omes):
        omeAccs = [acc for acc in accs if acc.startswith(ome + '_')]
        gff_list = gff2list(db[ome]['gff3'])
        grabAcc_res = [grabGffAccs(gff_list, omeAccs, ome)]

#        grabAcc_cmds.append([gff_list, omeAccs, ome])

 #   with mp.Pool(processes = cpus) as pool:
  #      grabAcc_res = pool.starmap(grabGffAccs, grabAcc_cmds)

    gffs = {}
    for res in grabAcc_res:
        gffs[res[1]] = res[0]

    return gffs


if __name__ == '__main__':

    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( 
        description = 'Inputs gff or database and new line delimitted file of accessions.'
        )
    parser.add_argument( '-a', '--accession' )
    parser.add_argument( '-i', '--input', help = 'File with accessions' )
    parser.add_argument( '-g', '--gff', help = 'GFF3 input' )
    parser.add_argument( '-c', '--column', default = 1, 
        help = 'Accssions column for -i (1 indexed). DEFAULT: 1' )
    parser.add_argument( '-o', '--ome', action = 'store_true', help = 'Output files by ome code' )
    parser.add_argument( '-d', '--database', default = masterDB(), help = 'mycodb DEFAULT: master' )
    parser.add_argument( '--cpu', type = int, default = mp.cpu_count(), \
        help = 'CPUs to parallelize accession searches. DEFAULT: all' )
    args = parser.parse_args()

    if args.input:
        input_file = formatPath( args.input )
        with open(input_file, 'r') as raw:
            accs = [x.rstrip().split('\t')[args.column-1] for x in raw]
    elif not args.accession:
        print('\nERROR: need input file or accession', flush = True)
        sys.exit( 1 )
    else:
        accs = [args.accession]

    if args.cpu < mp.cpu_count():
        cpu = args.cpu
    else:
        cpu = mp.cpu_count()

    db_path = formatPath( args.database )
    if not args.gff:
        db = mtdb( formatPath(args.database) )
        gff_lists = dbMain( db, accs, cpus = args.cpu )
    else:
        gff_path = formatPath( args.gff )
        gff_lists = gffMain(gffData, accs)

    if args.accession:
        print( list2gff(gff_lists[list(gff_lists.keys())[0]]).rstrip() , flush = True)
    elif args.ome:
        output = mkOutput(os.getcwd() + '/', 'acc2gff')
        for ome in gff_strs:
            if gff_lists[ome]:
                with open( output + ome + '.accs.gff3', 'w' ) as out:
                    out.write( list2gff(gff_lists[ome]) )
            else:
                eprint('ERROR: ' + ome + ' failed, no accessions retrieved', flush = True)
    else:
        out_str = ''
        for ome in gff_lists:
            if gff_lists[ome]:
                out_str += list2gff(gff_lists[ome]) + '\n'
            else:
                eprint('ERROR: ' + ome + ' does not have accession', flush = True)
        print(out_str)

    sys.exit( 0 )



#! /usr/bin/env python3

import os
import re
import sys
import argparse
import subprocess
import pandas as pd
import multiprocessing as mp
from StringIO import StringIO
from mycotools.lib.kontools import intro, outro, collect_files, multisub, findExecs
from mycotools.lib.dbtools import db2df
from mycotools.lib.biotools import dict2fa
from mycotools.extractHmmsearch import main as exHmm
from mycotools.acc2fa import main as acc2fa
from mycotools.extractHmmAcc import grabAccs, main as absHmm

def compileHmmCmd( db, hmmdb_path, output, ome_set = set() ):
    '''
    Inputs: mycotools db, hmmdb_path, output directory, set of omes to ignore.
    Outputs: tuples of arguments for hmmsearches
    '''

    cmd_tuples = [ ]
    for i, row in db.iterrows():
        if pd.isnull( row['faa'] ) or row['ome'] in ome_set:
            continue
        proteome_path = os.environ['MYCOFAA'] + '/' + row['faa']
        output_path = output + '/' + row['ome'] + '.out'
        cmd = ( 'hmmsearch', '-o', output_path, hmmdb_path, proteome_path ) 
        cmd_tuples.append( cmd )

    return cmd_tuples
    

def compileextractHmmCmd( db, args, output ):
    '''
    Inputs: mycotools db, argparse arguments, and output path
    Outputs: tuples of arguments for `runExHmm`
    '''

    cmd_tuples = [ ]
    for i, row in db.iterrows():
        if pd.isnull( row['faa'] ):
            continue
        hmmsearch_out = output + '/' + row['ome'] + '.out'
        cmd_tuples.append( ( args, hmmsearch_out, output ) )

    return cmd_tuples


def runExHmm( args, hmmsearch_out, output ):

    with open( hmmsearch_out, 'r' ) as raw:
        data = raw.read()
    ome = os.path.basename(hmmsearch_out).replace( '.out', '' )
    out_dict = None
    if len(data) > 100:
        out_dict = exHmm(data, args[0], args[1], args[2], args[3], header = False)
        out_dict = { x: [ome, out_dict[x][1]] for x in out_dict }
    else:
        print('\t' + ome + ' empty hmmsearch', flush = True)

    return out_dict
    

def compileacc2fa( db, output, q_dict ):

    cmd_tuples = []
    if not os.path.isdir(output):
        os.mkdir( output )
    fa_dict = { row['ome']: os.environ['MYCOFAA'] + '/' + row['faa'] \
        for i, row in db.iterrows() if not pd.isnull( row['faa'] ) }
    for q in q_dict:
        cmd_tuples.append( ( q_dict[q], 0, fa_dict, output + '/' + q + '.aa.fa' ) )

    return cmd_tuples


def runacc2fa( q_files, header, fa_dict, output ):

    fa_str = ''
    for ome in q_files:
        q = q_files[ome]
        df = pd.read_csv( StringIO(q), sep = '\t', header = None )
        fa_str += dict2fa(acc2fa( df, header, fa = fa_dict[ ome ] )) + '\n'
    
    with open( output, 'w' ) as out:
        out.write( fa_str )


def compileHmmalignCmd( output, accessions ):

    cmd_tuples = [ [], [] ]
    for acc in accessions:
        fa = output + '/fastas/' + acc + '.aa.fa'
        hmm = output + '/hmms/' + acc + '.hmm'
        align = output + '/aligns/' + acc + '.a2m'
        conv = output + '/aligns/' + acc + '.phylip'
        trim = output + '/trimmed/' + acc + '.trimal'
        if os.path.isfile( conv ):
            with open( conv, 'r' ) as raw:
                data = raw.read()
            if len( data ) > 10:
                continue
        hmmalign = ( 'hmmalign', '--outformat', 'A2m', '-o', align, hmm, fa )
        esl = ( 'esl-reformat', '-o', conv, 'phylip', align )
        cmd_tuples[0].append( hmmalign )
        cmd_tuples[1].append( esl )

    return cmd_tuples


def compileTrimalCmd( output, trimal = '', trimmed = None):

    trimal_args = trimal.split( ' ' )
    cmd_tuples = []
    aligns = collect_files( output + '/aligns', 'phylip' )
    if trimmed:
        trimmed = set( os.path.basename(x).replace('.trimal','') \
            for x in trimmed )
        aligns = [x for x in aligns if os.path.basename(x).replace('.phylip','') \
             not in trimmed]
    for align in aligns:
        trim = output + '/trimmed/' + \
            os.path.basename( align ).replace('.phylip','.trimal')
        args = ['trimal', '-in', align, '-out', trim]
        args.extend( trimal_args ) 
        cmd_tuples.append( tuple( args ) )

    return cmd_tuples


if __name__ == '__main__':

    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( description = 'Runs `hmmsearch` v each proteome in a' \
        + ' `.db`. For each query, extracts results and optionally outputs a compiled fasta' \
        + ', hmmalignment and/or trimmed alignment.' )
    parser.add_argument( '-i', '--input', required = True, help = 'Input database `.db`' )
    parser.add_argument( '-d', '--hmmdb', required = True, help = 'Hmm database `.hmm`' )
    parser.add_argument( '-o', '--output', help = 'User-specified output directory' )
    parser.add_argument( '-p', '--previous', help = 'Previous db2hmmsearch dir. ' + \
        'Be wary of incomplete outputs from interrupted runs' )
    parser.add_argument( '-c', '--cpu', type = int, \
        help = 'Processors to use. Default = All' )
    parser.add_argument( '-f', '--fasta', action = 'store_true', \
        help = 'Compile fastas for each query' )
    parser.add_argument( '-l', '--align', default = False, action = 'store_true', \
        help = 'Align fastas to original hmm and trim via `trimal`. Calls `-f`' )
    parser.add_argument( '-a', '--accession', action = 'store_true', default = False, \
        help = 'Extract accessions instead of queries (Pfam, etc). Requires `-f`' )
    parser.add_argument( '-b', '--best', type = int, help = '# top hits for each organism' \
        + ' Requires `-f`' )
    parser.add_argument( '-t', '--threshold', help = 'Query percent threshold (+/-). Requires `-f`' )
    parser.add_argument( '-e', '--evalue', help = 'E value threshold, e.g. ' \
        + '10^(-x) where x is the input. Requires `-f`' )
    parser.add_argument( '--trimal', default = '', help = \
        'User-specified trimAl commands in "", e.g.: "-strictplus -fasta"' )

    args = parser.parse_args()
    deps = ['hmmsearch']

    if args.previous:
        args.output = args.previous

    if not args.output:
        output = os.getcwd() + '/db2hmmsearch'
    else:
        output = os.path.abspath( args.output )
    if not os.path.isdir( output ):
        os.mkdir( output )

    if args.cpu and args.cpu < mp.cpu_count():
        cpu = args.cpu
    else:
        cpu = mp.cpu_count()

    trimal = args.trimal.replace('"','')

    if args.align:
        args.fasta = True
        deps.extend( ['hmmalign', 'trimal'] )
    else:
        trimal = None

    args_dict = { 
        'Database': args.input, 'Hmm database': args.hmmdb, 'Output': output,
        'Previous run': args.previous, 'CPUs': cpu, 'Fasta': args.fasta, 
        'Alignment': args.align, 'Accession': args.accession, 'Top hits': args.best, 
        'Threshold': args.threshold, 'E-value': args.evalue, 'trimAl args': trimal 
    }
    start_time = intro( 'db2hmmsearch', args_dict )
    findExecs( deps, exit = {deps})
    db = db2df( args.input )

#    main( args, db, output, cpu, previous = args.previous )
#def main( args, db, output, cpu, previous = False ):

    previous = args.previous
    ome_set, skip1, skip2 = set(), False, False
    if previous:
        print( '\nCompiling previous run' , flush = True)
        if os.path.isdir( output + '/fastas' ):
            fas = collect_files( output + '/fastas', 'aa.fa' )
            ranQueries = [ os.path.basename(fa).replace('.aa.fa','') for fa in fas ]
            with open( args.hmmdb, 'r' ) as hmmdb_raw:
                queries = grabAccs( hmmdb_raw.read() )
            if len( set( queries ).intersection( set( ranQueries ) ) ) == len( set( ranQueries ) ):
                skip1 = True
                for fa in fas:
                    with open( fa, 'r' ) as raw:
                        data = raw.read()
                    if len(data) < 10:
                        skip1 = False
                        break
            if skip1:
                if os.path.isdir( output + '/aligns' ):
                    phylips = collect_files( output + '/aligns', 'phylip' )
                    ranQueries = [os.path.basename(x).replace('.phylip','') \
                        for x in phylips ]
                    if len( set( queries ).intersection( set ( ranQueries ) \
                        ) ) == len( set( ranQueries ) ):
                        skip2 = True
                        for phy in phylips:
                            with open( phy, 'r' ) as raw:
                                data = raw.read()
                            if len(data) < 10:
                                skip2 = False
                                break
                
        if not skip1:
            if os.path.isfile( output + '/omes.tar.gz' ):
                if not os.path.isdir( output + '/omes' ):
                    tar_code = subprocess.call( 'tar -xzf ' + output + \
                        '/omes.tar.gz', stdout = subprocess.DEVNULL, \
                        stderr = subprocess.DEVNULL, shell = True )
            omes = collect_files( os.path.abspath( previous ) + '/omes', 'out' )
            ome_set = set( os.path.basename(x).replace('.out', '') for x in omes )
        else:
            print('\thmmsearch -> hits.fa DONE', flush = True)
            if skip2:
                print('\thits.fa -> phylip DONE', flush = True)

    if not skip1:
        print( '\nSubmitting omes to hmmsearch' , flush = True)
        if not os.path.isdir( output + '/omes' ):
            os.mkdir( output + '/omes' )
        hmmsearch_tuples = compileHmmCmd( db, args.hmmdb, output + '/omes', ome_set = ome_set )
        hmmsearch_codes = multisub( hmmsearch_tuples, processes = cpu - 1 )
        for code in hmmsearch_codes:
            if code['code'] != 0:
                print( '\tExit ' + str(code['code'], flush = True) + ' ' + \
                     os.path.basename( code['stdin'][-1] ) )
    
        print( '\nExtracting hmmsearch output' , flush = True)
        exHmm_args = [args.accession, args.best, args.threshold, args.evalue]
        exHmm_tuples = compileextractHmmCmd( db, exHmm_args, output + '/omes' )
        with mp.get_context( 'spawn' ).Pool(processes = cpu) as pool:
            hmmAligns = pool.starmap( runExHmm, exHmm_tuples )

        if args.fasta:
            subprocess.Popen(['tar', '-czf', output + "/omes.tar.gz", \
                output + '/omes', '--remove-files'], stdout = subprocess.DEVNULL, \
                stderr = subprocess.DEVNULL )
            print( '\nCompiling fastas' , flush = True)
            q_dict = {}
            for hit in hmmAligns:
                if hit:
                    for query in hit:
                        if query not in q_dict:
                            q_dict[query] = {}
                        ome = hit[query][0]
                        align = hit[query][1]
                        q_dict[query][ ome ] = align

            acc2fa_tuples = compileacc2fa( db, output + '/fastas', q_dict )
            with mp.get_context( 'spawn').Pool( processes = cpu ) as pool:
                pool.starmap( runacc2fa, acc2fa_tuples )        

    if args.align:
        if not skip2:
            print( '\nAligning fastas to hmms' , flush = True)
            if not os.path.isdir( output + '/hmms' ):
                os.mkdir( output + '/hmms' )
            with open( args.hmmdb, 'r' ) as hmmdb:
                hmm_dict = absHmm( hmmdb.read() )
            for accession in hmm_dict:
                with open( output + '/hmms/' + accession + '.hmm', 'w' ) as out:
                    out.write( hmm_dict[ accession ] )

            if not os.path.isdir( output + '/aligns' ):
                os.mkdir( output + '/aligns' )

            
            hmmAlign_tuples = compileHmmalignCmd( output, list(hmm_dict.keys() ) )
            hmmAlign_codes = multisub( hmmAlign_tuples[0], processes = cpu - 1 )
            esl_codes = multisub( hmmAlign_tuples[1], processes = cpu - 1 )
            for index in range( len (hmmAlign_codes) ):
                code = hmmAlign_codes[ index ]
                if code['code'] != 0:
                    print( '\thmmalign exit ' + str(code['code'], flush = True) + ' ' + \
                         os.path.basename( code['stdin'][-1] ) )
                else:
                    code = esl_codes[ index ]
                    if code['code'] != 0:
                        print( '\tesl-reformat exit ' + str(code['code'], flush = True) + ' ' + \
                            os.path.basename( code['stdin'][-1] ) )
                    os.remove( code['stdin'][-1] )
        
        print( '\nTrimming alignments' , flush = True)
        if not os.path.isdir( output + '/trimmed' ):
            os.mkdir( output + '/trimmed' )
        trimmed = None
        if args.previous:
            trimmed = collect_files( output + '/trimmed', 'trimal' )
        trimal_tuples = compileTrimalCmd( output, trimal = trimal, trimmed = trimmed )

        trimal_codes = multisub( trimal_tuples, processes = cpu - 1 )
        for code in trimal_codes:
            if code['code'] != 0:
                print( '\tExit ' + str(code['code'], flush = True) + ' ' + \
                     os.path.basename( code['stdin'][2] ) )

    outro( start_time )



#! /usr/bin/env python3

import os
import re
import sys
import time
import getpass
import argparse
import subprocess
import pandas as pd
from mycotools.lib.kontools import eprint, formatPath, outro, intro
from mycotools.lib.dbtools import loginCheck

def jgi_login( user, pwd ):
    '''Login via JGI's prescribed method by creating a cookie cache and 
    downloading JGI's sign-in file.'''

    null = os.path.expanduser( '~/.nullJGIdwnld' )
        
    login_cmd = subprocess.call(
        ['curl', 
        'https://signon.jgi.doe.gov/signon/create', '--data-urlencode', 'login=' + str(user), 
        '--data-urlencode', 'password=' + str(pwd), '-c', 'cookies', '-o', null],
        stdout = subprocess.PIPE, stderr = subprocess.PIPE
        )

    return login_cmd


def retrieveXML( ome, output ):
    '''Retrieve JGI xml file tree. First check if it already exists, if not then 
    download it using JGI's prescribed method. Then open the xml and check for 
    the common 'Portal does not exist' error. If so, report.'''

    if os.path.exists( output + '/' + str(ome) + '.xml' ):
        with open( output + '/' + str(ome) + '.xml', 'r' ) as xml_raw:
            xml_data = xml_raw.read()
        if xml_data == 'Portal does not exist':
            print( '\tERROR: `' + ome + ' not in JGIs `organism` database.' , flush = True)
            xml_cmd = 1
            os.remove( output + '/' + ome + '.xml' )
        else:
            xml_cmd = -1

    else:
        xml_cmd = subprocess.call( [
            'curl', 
            "https://genome.jgi.doe.gov/portal/ext-api/downloads/get-directory?organism=" +
            str(ome), '-b', 'cookies', '-o', output + '/' + str(ome) + ".xml" ],
            stdout = subprocess.PIPE, stderr = subprocess.PIPE )
        if xml_cmd != 0:
            print('\tERROR: `xml` directory ' + ome + ' - `curl` error: ' + str(xml_cmd), flush = True)

    if xml_cmd == 0:
        with open( output + '/' + ome + '.xml', 'r' ) as xml_raw:
            xml_data = xml_raw.read()
        if xml_data == 'Portal does not exist':
            print( '\tERROR: `' + ome + ' not in JGIs `organism` database.' , flush = True)
            xml_cmd = 1
            os.remove( output + '/' + ome + '.xml' )

    return xml_cmd


def checkMatches( match1, match2, match3, match4, xml_data, file_type):

    findall, preexisting = False, None
    matches = match1.search(xml_data)
    if matches is None or 'get_tape_file' in matches[0]:
        matches = match2.search(xml_data)
        if matches is None or 'get_tape_file' in matches[0]:
            if match3:
                matches = match3.search(xml_data)
                if matches is None or 'get_tape_file' in matches[0]:
                    if match4:
                        matches = match4.search(xml_data)
            if matches is None or 'get_tape_file' in matches[0]:
                matches = match1.findall(xml_data)
                matches = [ x for x in matches if 'get_tape_file' not in x ]
                findall = True
                if not matches: # changed from if to if not 20210624
                    matches = match2.findall(xml_data)
                    matches = [ x for x in matches if 'get_tape_file' not in x ]
                    if not matches and match3:
                        matches = match3.findall(xml_data)
                        matches = [ x for x in matches if 'get_tape_file' not in x ]
                        if not matches and match4:
                            matches = match4.findall(xml_data)
                            matches = [x for x in matches if 'get_tape_file' not in x]
                if len(matches) > 1:
                    matches = matches[1]
                else:
                    if file_type != 'gff3':
                        print('\t\tERROR: no ' + file_type + ' detected.', flush = True)
                    matches = None
                    preexisting = True

    return matches, preexisting, findall


def JGIdwnld( ome, file_type, output, masked = True ):
    '''Download JGI files. For each type of file, use regular expressions to 
    gather the URL from the file. Grab the md5checksum if possible from the 
    xml as well. Create arbitrary values for md5 and curl_cmd. If the download file
    already exists, then run an md5 checksum if an md5 value exists in the xml.

    If the md5 does not match then open and check for typical errors. If those 
    errors exist, obtain the alternative download URL from the xml. If the md5 
    matches, pass through the rest of the function.

    If there is no download md5, then check to see if the file is greater than 10 
    lines as a proxy to make sure that the file isn't empty/blatantly wrong. If it 
    passes this test, change the values to not enter the while loop at the end of 
    the function.

    The while loop following the file exists error allows for 3 attempts. If a file 
    is downloaded it will check its md5 using the xml reference. If it fails, it will 
    proceed via the error checking above, wait a minute, and reattempt. If there is no 
    download md5 from the xml it will proceed via the error checking above as well. If it 
    passes the line count check, it will exit the loop - otherwise, it will attempt to 
    gather a new URL and restart after another minute wait. This is an unfortunate
    circumnavigation of JGI's lack of a consistent maximum ping / time before 
    booting.'''


    prefix = 'https://genome.jgi.doe.gov'
    with open( output + '/xml/' + ome + '.xml', 'r') as xml:
        data = xml.read()

    gffMatch1 = re.compile(r'filename\="([^ ]+\_GeneCatalog_genes_\d+\.gff\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
    gffMatch2 = re.compile(r'filename\="([^ ]+\_GeneCatalog_genes_\d+\.gff\.gz)".*?url\="([^ ]+)"')
    match3, match4 = None, None
    if file_type == 'gff':
        match1 = gffMatch1
        match2 = gffMatch2
    elif file_type == 'fna':
        if masked:
            match1 = re.compile(r'filename\="(' + ome + r'\_AssemblyScaffolds_Repeatmasked\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
            match2 = re.compile(r'filename\="(' + ome + r'\_AssemblyScaffolds_Repeatmasked\.fasta\.gz)".*?url\="([^ ]+)"')
            match3 = re.compile(r'filename\="([^ ]+?maskedAssembly\.gz)".*? url="([^ ]*?maskedAssembly.gz)".*? md5\="([\w\d]+)"')
        else:
            match1 = re.compile(r'filename\="(' + ome + r'\_AssemblyScaffolds\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
            match2 = re.compile(r'filename\="(' + ome + r'\_AssemblyScaffolds\.fasta\.gz)".*?url\="([^ ]+)"')
            match3 = re.compile(r'filename\="([^ ]+?AssembledScaffolds.*?\.gz)".*? url="(.*?AssembledScaffolds[^ ]*?\.gz)".*? md5\="([\w\d]+)"')
            match4 = re.compile(r'filename\="([^ ]+?AssembledScaffolds.*?\.gz)".*? url="(.*?AssembledScaffolds[^ ]*?\.gz)"')

    elif file_type == 'faa':
        match1 = re.compile(r'filename\="([^ ]+\_GeneCatalog_proteins_\d+\.aa\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
        match2 = re.compile(r'filename\="([^ ]+\_GeneCatalog_proteins_\d+\.aa\.fasta\.gz)".*?url\="([^ ]+)"')
        match3 = re.compile(r'filename\="([^ ]+?\.FilteredModels[^ ]+\.proteins\.[^ ]*?\.gz)".*? url\="([^ ]*?)".*? md5\="([\w\d]+)"')
        match4 = re.compile(r'filename\="([^ ]+?\.FilteredModels[^ ]+\.proteins\.[^ ]*?\.gz)".*? url\="([^ ]*?)"')

    elif file_type == 'gff3':
        match1 = re.compile(r'filename\="([^ ]+\_GeneCatalog\_\d+\.gff3\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
        match2 = re.compile(r'filename\="([^ ]+\_GeneCatalog\_\d+\.gff3\.gz)".*?url\="([^ ]+)"')
        match3 = re.compile(r'filename\="([^ ]+?\.FilteredModels3\.gff3\.gz)".*?url\="([^ ]*?)".*?md5\="([\d\w]+)"')
        match4 = re.compile(r'filename\="([^ ]+?\.FilteredModels3\.gff3\.gz)".*?url\="([^ ]*?)"')

    elif file_type == 'transcript':
        match1 = re.compile(r'filename\="([^ ]+_GeneCatalog_transcripts_\d+\.nt\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
        match2 = re.compile(r'filename\="([^ ]+\_GeneCatalog_transcripts_\d+\.nt\.fasta\.gz)".*?url\="([^ ]+)"')

    elif file_type == 'AllTranscript':
        match1 = re.compile(r'filename\="([^ ]+\_all_transcripts_\d+\.nt\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
        match2 = re.compile(r'filename\="([^ ]+\_all_transcripts_\d+\.nt\.fasta\.gz)".*?url\="([^ ]+)"')

    elif file_type == 'est':
        match1 = re.compile(r'filename\="([^ ]+\_EST_\d+_cluster_consensi\.fasta\.gz)".*?url\="([^ ]+)".*?md5\="([^ ]+)"')
        match2 = re.compile(r'filename\="([^ ]+\_EST_\d+_cluster_consensi\.fasta\.gz)".*?url\="([^ ]+)"')

    check = 1
    matches, preexisting, findall = checkMatches( match1, match2, match3, match4, data, file_type )
    if not matches and file_type == 'gff3':
        file_type = 'gff'
        matches, preexisting, findall = checkMatches( gffMatch1, gffMatch2, match3, match4, data, file_type )
    if matches:

        md5 = 420
        attempt = 0
        curl_cmd = 420

        if findall:
            dwnld_url = prefix + matches[1].replace( '&amp;', '&' )
        else:
            dwnld_url = prefix + matches[2].replace( '&amp;', '&' )

#        print('\t\t' + dwnld_url, flush = True)
        if match1.search(data):
            dwnld_md5 = match1.search(data)[3]
        else:
            dwnld_md5 = None
        dwnld = os.path.basename( dwnld_url )

        if os.path.exists( output + '/' + file_type + '/' + dwnld):
            if dwnld_md5:
                md5_cmd = subprocess.run( [
                        'md5sum', output + '/' + file_type + '/' + dwnld], 
                        stdout = subprocess.PIPE )
                md5_res = md5_cmd.stdout.decode( 'utf-8' )
                md5_find = re.search(r'\w+', md5_res)
                md5 = md5_find[0]

            if md5 == dwnld_md5:
                curl_cmd = 0
                check = dwnld
                preexisting = True
            else:
                while True:
                    try:
                        with open(output + '/' + file_type + '/' + dwnld, 'r') \
                            as dwnld_data_raw:
                            dwnld_data = dwnld_data_raw.read()
                        if re.search( '307 Temporary Redirect', dwnld_data ):
                            print('\t\t' + dwnld + ' link has moved. Trying a different link.', flush = True)
                            matches1 = re.search( 
                                    r'url\="(\/portal\/' + ome + r'\/download\/' + \
                                    dwnld + r')', data)
                            if matches1:
                                dwnld_url = prefix + matches1[1].replace( '&amp;', '&' )
                                print('\t\t' + dwnld_url, flush = True)
                            break
                        else:
                            if not dwnld_md5:
                                check_size = subprocess.run( 
                                        ['wc', '-l', output + '/' + file_type + '/' + dwnld], 
                                        stdout = subprocess.PIPE 
                                        )
                                check_size_res = check_size.stdout.decode( 'utf-8' )
                                print('\t\tFile exists - no md5 to check.', flush = True)
                                check_size_find = re.search(r'\d+', check_size_res)
                                size = check_size_find[0]
                                if int(size) < 10:
                                    print('\t\tInvalid file size.', flush = True)
                                else:
                                    md5 = None
                            else:
                                print('\t\tFile exists, but md5 does not match.', flush = True)
                            break
                    except UnicodeDecodeError:
                        if not dwnld_md5:
                            print('\t\tFile exists - no md5 to check.', flush = True)
                            check_size = subprocess.run( 
                                    ['wc', '-l', output + '/' + file_type + '/' + dwnld], 
                                    stdout = subprocess.PIPE )
                            check_size_res = check_size.stdout.decode( 'utf-8' )
                            check_size_find = re.search(r'\d+', check_size_res)
                            size = check_size_find[0]
                            if int(size) < 10:
                                print('\t\tInvalid file size.', flush = True)
                            else:
                                md5 = None
                                preexisting = True
                                check = dwnld
                        else:
                            print('\t\tFile exists, but md5 does not match.', flush = True)
                        break

        while md5 != dwnld_md5 and curl_cmd != 0 and attempt < 3:
            attempt += 1
            curl_cmd = subprocess.call( 
                    ['curl', dwnld_url, '-b', 'cookies', '-o', 
                    output + '/' + file_type + '/' + dwnld], 
                    stdout = subprocess.PIPE, stderr = subprocess.PIPE 
                    )

            if curl_cmd == 0:
                check = dwnld
                
                if dwnld_md5:
                    md5_cmd = subprocess.run( 
                            ["md5sum", output + '/' + file_type + '/' + dwnld], 
                            stdout = subprocess.PIPE 
                            )
                    md5_res = md5_cmd.stdout.decode( 'utf-8' )
                    md5_find = re.search(r'\w+', md5_res)
                    md5 = md5_find[0]

                if not dwnld_md5:
                        try:
                            with open(output + '/' + file_type + '/' + dwnld, 'r') \
                                as dwnld_data_raw:
                                dwnld_data = dwnld_data_raw.read()
                            if re.search( '307 Temporary Redirect', dwnld_data ):
                                print(
                                    '\t\t\t' + dwnld + ' link has moved. ' + \
                                    'Trying a different link.'
                                    )
                                matches1 = re.search( 
                                        r'url\="(\/portal\/' + ome + r'\/download\/' + \
                                        dwnld + r')', data
                                        )
                                if matches1:
                                    dwnld_url = prefix + matches1[1].replace( '&amp;', '&' )
                                    print('\t\t' + dwnld_url, flush = True)
                                else:
                                    print('\t\t\t\tNo valid alternative link. Skip.', flush = True)
                                    attempt = 4
                                    break
                            else:
                                print('\t\tFile exists - no md5 to check.', flush = True)
                                check_size = subprocess.run( 
                                        "wc -l " + output + '/' + file_type + '/' + dwnld, 
                                        stdout = subprocess.PIPE 
                                        )
                                check_size_res = check_size.stdout.decode( 'utf-8' )
                                check_size_find = re.search(r'\d+', check_size_res)
                                size = check_size_find[0]
                                if int(size) < 10:
                                    print('\t\tInvalid file size. Attempt ' + str(attempt), flush = True)
                                else:
                                    md5 = None
                                    break
                        except UnicodeDecodeError:
                            pass
                        time.sleep( 60 )

                elif md5 != dwnld_md5 and attempt == 1:
                    print('\t\tERROR: md5 does not match JGI. Attempt ' + str(attempt), flush = True)
                    curl_cmd = -1
                    check = 2
                    while True:
                        try:
                            with open(output + '/' + file_type + '/' + dwnld, 'r') \
                                as dwnld_data_raw:
                                dwnld_data = dwnld_data_raw.read()
                            if re.search( '307 Temporary Redirect', dwnld_data ):
                                print(
                                    '\t\t\t' + dwnld + ' link has moved. ' + \
                                    'Trying a different link.'
                                )
                                matches1 = re.search( r'url\="(\/portal\/' + \
                                        ome + r'\/download\/' + dwnld + r')', data)
                                if matches1:
                                    dwnld_url = prefix + matches1[1].replace( '&amp;', '&' )
                                    print('\t\t' + dwnld_url, flush = True)
                                else:
                                    print('\t\t\t\tNo valid alternative link. Skip.', flush = True)
                                    attempt = 4
                            break
                        except UnicodeDecodeError:
                            pass
                            break
                        time.sleep( 60 )
                elif md5 != dwnld_md5 and attempt == 2:
                    print('\t\tERROR: md5 does not match JGI. Attempt ' + str(attempt), flush = True)
                    curl_cmd = -1
                    matches1 = re.search( 
                        r'url\="(\/portal\/' + ome + r'\/download\/' + dwnld + r')', data
                        )
                    if matches1:
                        dwnld_url = prefix + matches1[1].replace( '&amp;', '&' )
                        print('\t\t' + dwnld_url, flush = True)
                    time.sleep( 60 )
                    check = 2
                elif md5 != dwnld_md5:
                    print('\t\tERROR: md5 does not match JGI. Attempt ' + str(attempt), flush = True)
                    check = 2
            else:
                print(
                    '\t\tERROR: Failed to retrieve ' + file_type + '. `curl` error: ' \
                    + str(curl_cmd) + '\nAttempt' + str(attempt)
                    )
                check = 2

        if attempt == 3:
            if md5 != dwnld_md5:
                print('\t\tExcluding from database - potential failure', flush = True)
                curl_cmd = 0
            if curl_cmd != 0:
                print('\t\tFile failed to download', flush = True)

#    else:
 #       print('\t\tfailed', flush = True)

    return check, preexisting, file_type



def main( 
    df, output, user, pwd, assembly = True, proteome = False, gff3 = True, 
    gff = False, transcript = False, est = False, masked = True
    ):

#    pd.options.mode.chained_assignment = None  # default='warn'
    if not 'assembly_acc' in df.columns:
        if len( df.columns ) != 1:
            eprint( '\nInvalid input. No assembly_acc column and more than one column.' , flush = True)
        else:
            ome_col = list(df.columns)[0]
    else:
        ome_col = 'assembly_acc'

    eprint('\nLogging into JGI', flush = True)
    login_attempt = 0
    while jgi_login( user, pwd ) != 0 and login_attempt < 5:
        eprint('\nJGI Login Failed. Attempt: ' + str(login_attempt), flush = True)
        time.sleep( 5 )
        login_attempt += 1
        if login_attempt == 5:
            eprint('\nERROR: Failed login 5 attempts.', flush = True)
            sys.exit( 100 )

    if not os.path.exists( output + '/xml' ):
        os.mkdir( output + '/xml' )
# perhaps add a counter here, but one that checks if it is actually querying jgi
    print( '\nRetrieving `xml` directories ...' , flush = True)
    ome_set, count = set(), 0
    for i,row in df.iterrows():
        error_check = retrieveXML( row[ome_col], output + '/xml' )
        if error_check > 0:
            ome_set.add( row[ome_col] )
        if error_check != -1:
            time.sleep( 0.1 )

    eprint( '\nDownloading JGI files\n\tMaximum rate: 1 genome/min' , flush = True)
    if len( df ) > 60:
        eprint( '\tMinimum time: ' + str(len(df)) / 60 + ' hours' , flush = True)
    else:
        eprint( '\tMinimum time: ' + str(len(df)) + ' minutes' , flush = True)
    
    dwnlds = []
    if assembly:
        dwnlds.append( 'fna' )
    if proteome:
        dwnlds.append( 'faa' )
    if gff3:
        dwnlds.append( 'gff3' )
    if gff:
        dwnlds.append( 'gff' )
    if transcript:
        dwnlds.append( 'transcript' )
#        dwnlds.append( 'AllTranscript' )
    if est:
        dwnlds.append( 'est' )

    for typ in dwnlds:
        if not os.path.isdir( output + '/' + typ ):
            os.mkdir( output + '/' + typ )
        if typ == 'gff3':
            if not os.path.isdir( output + '/gff' ):
                os.mkdir( output + '/gff')

    preexisting = True
    for i, row in df.iterrows():
        ome = row[ome_col]
        if not preexisting:
            time.sleep( 60 )
        if ome not in ome_set:
            jgi_login( user, pwd )
            if 'ome' in row.keys():
                eprint( '\t' + row['ome'] + '\t' + ome , flush = True)
            else:
                eprint( '\t' + ome , flush = True)     
            for typ in dwnlds:
                check, preexisting, new_typ = JGIdwnld(
                    ome, typ, output, masked = masked
                )
                if type(check) != int:
                    if new_typ == 'gff':
                        df.loc[i, 'jgi_gff2_path'] = output + '/' + new_typ \
                            + '/' + check
                    else:
                        df.loc[i, new_typ + '_path'] = output + '/' + new_typ + \
                            '/' + check
                    check = os.path.basename( os.path.abspath( check ) )
                elif type(check) == int:
                    ome_set.add(row[ome_col])
                eprint('\t\t' + new_typ + ': exit status ' + str(check), flush = True)
        else:
            eprint( '\t' + ome + ' failed.' , flush = True)

    if os.path.exists( 'cookies' ):
        os.remove( 'cookies' )
    if os.path.exists( os.path.expanduser( '~/.null' )):
        os.remove( os.path.expanduser( '~/.null' ))

    if 'gff3' in df.columns:
        del df['gff3']
    if 'faa' in df.columns:
        del df['faa']
    if 'fna' in df.columns:
        del df['fna']

    return df, ome_set


if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = \
        "Imports table/database with JGI `assembly_acc` column and downloads assembly, proteome, gff, " + \
        'and/or gff3. This script supports rerunning/continuing previous runs in the same directory. ' + \
        'JGI has stringent, nondefined ping limits, so file downloads are limited to 1 per minute.' )
    parser.add_argument( '-i', '--input', required = True, help = 'Genome code or table with `assembly_acc` column of JGI ome codes' )
    parser.add_argument( '-a', '--assembly', default = False, action = 'store_true', \
        help = 'Download assemblies.' )
    parser.add_argument( '-p', '--proteome', default = False, action = 'store_true', \
        help = 'Download proteomes.' )
    parser.add_argument( '-g', '--gff', default = False, action = 'store_true', \
        help = 'Download gffs.' )
    parser.add_argument( '-g3', '--gff3', default = False, action = 'store_true', \
        help = 'Download gff3s.' )
    parser.add_argument( '-t', '--transcript', default = False, action = 'store_true', \
        help = 'Download transcripts.' )
    parser.add_argument( '-e', '--est', default = False, action = 'store_true', \
        help = 'Download EST data.' )
    parser.add_argument( '--nonmasked', default = False, action = 'store_true', \
        help = "Download nonmasked assemblies. Requires -a" )
    parser.add_argument( '-o', '--output', default = os.getcwd(), help = 'Output dir' )
    args = parser.parse_args()

    if args.nonmasked:
        args.assembly = True

    if not args.assembly and not args.proteome \
        and not args.gff and not args.transcript \
        and not args.est and not args.gff3:
        eprint( '\nERROR: You must choose at least one download option.' , flush = True)

    ncbi_email, ncbi_api, user, pwd = loginCheck( ncbi = False )
#    user = input( 'JGI username: ' )
 #   pwd = getpass.getpass( prompt='JGI Login Password: ' )

    args_dict = {
            'JGI Table': args.input,
            'Assemblies': args.assembly,
            'RepeatMasked': not args.nonmasked,
            'Proteomes': args.proteome,
            ".gff's": args.gff,
            ".gff3's": args.gff3,
            'Transcripts': args.transcript,
            'EST': args.est
            }

    start_time = intro( 'Download JGI files', args_dict )
   
    if os.path.isfile(args.input): 
        with open(args.input, 'r') as raw:
            for line in raw:
                if 'assembly_acc' in line.rstrip().split('\t'):
                    df = pd.read_csv( args.input, sep = '\t', index_col = None)
                else:
                    df = pd.read_csv(args.input, sep='\t', header = None)
                break
    else:
        df = pd.DataFrame({'assembly_acc': [args.input]})

    output = formatPath(args.output)

    jgi_df = main( 
        df, output, user, pwd, args.assembly, args.proteome, 
        args.gff3, args.gff, args.transcript, args.est, not args.nonmasked
        )
    df.to_csv( os.path.normpath(args.input) + '_jgiDwnld', sep = '\t', index = False )

    outro( start_time )




#! /usr/bin/env python3

import os
import re
import sys
import copy
import json
import time
import base64
import getpass
import hashlib
import datetime
import numpy as np
from Bio import Entrez
from urllib.error import HTTPError
from io import StringIO
from collections import defaultdict
from mycotools.lib.kontools import collect_files, eprint, formatPath


class mtdb(dict):
    '''
    MycotoolsDB (mtdb) class. Designed to support high throughput "pandas-like"
    interface without the overhead of pandas.
    '''
    columns = [
       'ome', 'genus', 'species', 'strain', 'taxonomy',
       'version', 'source', 'biosample', 'assembly_acc',
       'acquisition_date', 'published', 'fna', 'faa', 'gff3'
       ]

    # NEED a detect index feature for adding dicts in with alternative indices
    def __init__(self, db = None, index = None):
        self.columns = [
            'ome', 'genus', 'species', 'strain', 'taxonomy',
            'version', 'source', 'biosample', 'assembly_acc',
            'acquisition_date', 'published', 'fna', 'faa', 'gff3'
            ]
        if not db:
            if not index:
                super().__init__({x: [] for x in self.columns})
            else:
                super().__init__()
        elif isinstance(db, dict):
            if not index:
                if all(isinstance(db[x], list) for x in self.columns):
                    super().__init__(db)
                    self.index = None
            else:
                try:
                    if set(db[list(db.keys())[0]].keys()).union({index}) \
                        == set(self.columns):
                        super().__init__(db)
                except AttributeError:
                    if set(db[list(db.keys())[0]][0]).union({index}) \
                        == set(self.columns):
                        super().__init__(db)
        else:
            super().__init__(mtdb.db2df(self, db))
        self.index = index

    def pd2mtdb(df): # legacy integration
        df = df.fillna('')
        db = mtdb({
            x: list(df[x]) for x in mtdb.columns
            })
        return db
        
    def db2df(self, db_path):
        df = defaultdict(list)
        with open(formatPath(db_path), 'r') as raw:
            data = [x.rstrip().split('\t') for x in raw if not x.startswith('#')]
        if len(data[0]) == 16: # legacy conversion TO BE DEPRECATED
            eprint('\tWARNING: Legacy MycotoolsDB format will be removed in the future.', flush = True)
            columns = [
                'ome', 'genus', 'species', 'strain', 'version',
                'biosample', 'fna', 'faa', 'gff3', 'taxonomy', 'ecology',
                'eco_conf',
                'source', 'published', 'assembly_acc', 'acquisition_date'
                ]
        else:
            columns = self.columns
        for entry in data:
            [df[c].append('') for c in columns] # add a blank entry to each
            # column
            for i, d in enumerate(entry):
                df[columns[i]][-1] = d
            df['taxonomy'][-1] = read_tax(
                df['taxonomy'][-1]
                )
            df['taxonomy'][-1]['genus'] = df['genus'][-1]
            df['taxonomy'][-1]['species'] = \
                df['genus'][-1] + ' ' + df['species'][-1]
        if len(data[0]) == 16: # LEGACY conversion to be deprecated
            del df['ecology']
            del df['eco_conf']
        for i, ome in enumerate(df['ome']): 
            # if malformatted due to decreased entries in some lines, this will raise an IndexError
            if not df['fna'][i] or df['fna'][i] == ome + '.fa': # abbreviated line w/o file coordinates
               df['fna'][i] = os.environ['MYCOFNA'] + '/' + ome + '.fa'
               df['faa'][i] = os.environ['MYCOFAA'] + '/' + ome + '.aa.fa'
               df['gff3'][i] = os.environ['MYCOGFF3'] + '/' + ome + '.gff3'
            else: # has file coordinates
               df['fna'][i] = formatPath(df['fna'][i])
               df['faa'][i] = formatPath(df['faa'][i])
               df['gff3'][i] = formatPath(df['gff3'][i])

        return df


    def df2db(self, db_path = None, headers = False):
        df = copy.copy(self)
        df = df.reset_index()
        output = mtdb(
            {k: v for k,v in sorted(self.set_index('ome').items(), key = lambda x: x[0])},
            index = 'ome'
            ) 
        # does this work if its not an inplace change
        paths = {
            'faa': [os.environ['MYCOFAA'] + '/', '.aa.fa'],
            'fna': [os.environ['MYCOFNA'] + '/', '.fa'],
            'gff3': [os.environ['MYCOGFF3'] + '/', '.gff3']
            }
        if db_path:
            with open(db_path, 'w') as out:
                if headers:
                    out.write('\t'.join(self.column)+ '\n')
                for ome in output:
                    for file_type in ['fna', 'faa', 'gff3']:
                        output[ome][file_type] = output[ome][file_type].replace(
                            paths[file_type][0] + ome + paths[file_type][1], ''
                            ) # abbreviate when possible
                    if output[ome]['taxonomy']:
                        output[ome]['taxonomy'] = json.dumps(output[ome]['taxonomy'])
                    else:
                        output[ome]['taxonomy'] = '{}'
                    out.write(
                        ome + '\t' + \
                        '\t'.join([str(output[ome][x]) for x in output[ome]]) + '\n'
                        )
        else:
            if headers:
                print(
                    '\t'.join(self.column), flush = True
                    )

            for ome in output:
                for file_type in ['fna', 'faa', 'gff3']:
                    output[ome][file_type] = output[ome][file_type].replace(
                        paths[file_type][0] + ome + paths[file_type][1], ''
                        ) # abbreviate when possible
                output[ome]['taxonomy'] = json.dumps(output[ome]['taxonomy'])
                print(ome + '\t' + \
                    '\t'.join([str(output[ome][x]) for x in output[ome]]), flush = True
                    )


    def set_index(self, column = 'ome', inplace = False):
        data, retry, error, df, columns = {}, bool(column), False, copy.copy(self), copy.copy(self.columns)
        if not column:
            return df.reset_index()
        while retry:
            oldCol = set()
            try:
                columns.pop(columns.index(column))
            except (ValueError, IndexError) as e:
                df = df.reset_index() #will this actually reset the index
                columns.pop(columns.index(column))
            try:
                df[column]
            except KeyError:
                df = df.reset_index()
            if len(df[column]) == 0:
                return mtdb({}, index = column)
            if column in \
                {'assembly_acc', 'ome', 'fna', 'gff3', 'faa'}: # if it's unique
                for i, v in enumerate(df[column]):
                    data[v] = {}
                    for head in columns:
                        data[v][head] = df[head][i]
            else:
                for i, v in enumerate(df[column]):
                    if v not in data:
                        data[v] = []
                    data[v].append({})
                    for head in columns:
                        try:
                            data[v][-1][head] = df[head][i]
                        except KeyError: # if an index exists
                            if error:
                                eprint('\nERROR: invalid column', flush = True)
                                return self
                            df = df.reset_index()
                            error = True
                            break

            retry = False
        return mtdb(data, column)

    def reset_index(self):
        df = copy.copy(self)
        if df.index:
            data = {x: [] for x in mtdb().columns}
            data[df.index] = list(df.keys())
            if df.index in {'assembly_acc', 'ome', 'biosample', 'fna', 'gff3', 'faa'}:
                for key in df:
                    for otherCol in df[key]:
                        data[otherCol].append(df[key][otherCol])
            else:
                for key in df:
                    for v in df[key]:
                        for otherCol in v:
                            data[otherCol].append(v[otherCol])
            return mtdb(data, index = None)
        else:
            return df

    def append(self, info = {}):
#        if any(x not in set(self.columns) for x in info):
 #           raise KeyError('Invalid keys: ' + str(set(info.keys()).difference(set(self.columns))))
        index = self.index
        df = copy.copy(self)
        df = df.reset_index()
        info = {**info, **{k: None for k in set(self.columns).difference(set(info.keys()))}}
        for key in self.columns:
            df[key].append(info[key])
        return df.set_index(index)
        
              
                
def getLogin( ncbi, jgi ):

    ncbi_email, ncbi_api, jgi_email, jgi_pwd = None, None, None, None
    print(flush = True)
    if ncbi:
        ncbi_email = input( 'NCBI email: ' )
        ncbi_api = getpass.getpass( prompt = 'NCBI api key (blank if none): ' )
    if jgi:
        jgi_email = input( 'JGI email: ' )
        jgi_pwd = getpass.getpass( prompt = 'JGI password (required): ' )
    print(flush = True)

    return ncbi_email, ncbi_api, jgi_email, jgi_pwd


def loginCheck( info_path = '~/.mycodb', ncbi = True, jgi = True ):

    salt = b'D9\x82\xbfSibW(\xb1q\xeb\xd1\x84\x118'
    #NEED to make this store a password
    if os.path.isfile( formatPath(info_path) ):
        from cryptography.fernet import Fernet
        from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
        from cryptography.hazmat.backends import default_backend
        kdf = PBKDF2HMAC(
            algorithm=hashlib.sha256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=default_backend()
        )
        print(flush = True)
        if sys.stdin.isatty():
            hash_pwd = getpass.getpass( prompt = 'MycoDB login password (stdin allowed): ' )
        else:
            hash_pwd = sys.stdin.readline().rstrip()
        key = base64.urlsafe_b64encode(kdf.derive(hash_pwd.encode('utf-8')))
        fernet = Fernet( key )
#        with open(formatPath(info_path) + '/.key', 'rb') as raw_key:
 #           fernet = Fernet(raw_key)
        with open(formatPath(info_path), 'rb') as raw_file:
            data = raw_file.read()
        decrypted = fernet.decrypt(data)
        data = decrypted.decode('UTF-8').split('\n')
        if len(data) != 4:
            eprint('BAD PASSWORD FILE. Delete ~/.mycodb to reset.', flush = True)
            sys.exit(8)
        ncbi_email = data[0].rstrip()
        ncbi_api = data[1].rstrip()
        jgi_email = data[2].rstrip()
        jgi_pwd = data[3].rstrip()
    else:
        ncbi_email, ncbi_api, jgi_email, jgi_pwd = getLogin(ncbi, jgi)
        if ncbi and jgi:
            hash_check = input( 'Would you like to encrypt your login ' + \
                'information to ' + info_path + ' [Y/n]: ' )
        else:
            hash_check = 'n'
        if hash_check not in {'n', 'N'}:
            from cryptography.fernet import Fernet
            from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
            from cryptography.hazmat.backends import default_backend
            kdf = PBKDF2HMAC(
                algorithm=hashlib.sha256(),
                length=32,
                salt=salt,
                iterations=100000,
                backend=default_backend()
            )
            print(flush = True)
            hash_pwd, hash_check = True, False
            while hash_pwd != hash_check:
                if hash_pwd != True:
                    eprint('ERROR: passwords do not match', flush = True)
                hash_pwd = getpass.getpass( prompt = 'New MycoDB login password: ' )
                hash_check = getpass.getpass( prompt = 'Confirm password: ' )

            key = base64.urlsafe_b64encode(kdf.derive(hash_pwd.encode('utf-8')))
            fernet = Fernet( key )
            out_data = ncbi_email + '\t' + ncbi_api + '\t' + jgi_email + '\t' + jgi_pwd
            encrypt_data = fernet.encrypt(out_data)
            with open( formatPath(info_path), 'wb' ) as out:
                out.write(encrypt_data)
#
    return ncbi_email, ncbi_api, jgi_email, jgi_pwd


def genConfig( 
    branch = 'stable', forbidden = '$MYCODB/log/forbidden.tsv',
    repo = "https://gitlab.com/xonq/mycodb", 
    rogue = False, nonpublished = False
    ):

    config = { 
        'forbidden': forbidden,
        'repository': repo, 
        'branch': branch,
        'nonpublished': nonpublished,
        'rogue': rogue,
    }

    return config



# opens a `log` file path to read, searches for the `ome` code followed by a whitespace character, and edits the line with `edit` 
def log_editor( log, ome, edit ):

    with open( log, 'r' ) as raw:
        data = raw.read()

    if re.search( ome + r'\s', data ):
        new_data = re.sub( ome + r'\s.*', edit, data )
    else:
        data = data.rstrip()
        data += '\n'
        new_data = data + edit

    with open( log, 'w' ) as towrite:
        towrite.write(new_data)


def readLog( log, columns = '', sep = '\t' ):

    log_dict = {}
    with open( log, 'r' ) as raw:
        data = raw.read()
    dataLines = [ x.split( sep ) for x in data.split( '\n' ) ]

    if type( columns ) is str:
        if dataLines[0][0].startswith( '#' ):
            dataLines[0][0] = re.sub( r'^\#+', '', dataLines[0][0] )
        columns = [ head for head in dataLines[0] ]
        del dataLines[0]
    elif type( columns ) is int:
        x = 0
        columns = [ x + 1 for y in dataLines[0] ]

    for line in dataLines:
        log_dict[ line[0] ] = { columns[ x ]: line[ x ] for x in range( 1, len( line ) ) }

    return log_dict


def masterDB(path = '$MYCODB'):
    """Acquire the path of the master database by searching $MYCODB for a file
    with a basename that starts with a date string %Y%m%d."""

    path = path.replace('$', '')
    try:
        full_path = os.environ[path]
    except KeyError: # $MYCODB not initialized
        return None

    files = collect_files(full_path, 'mtdb')
    basenames = [os.path.basename(x) for x in files]
    dates = [x.replace('.mtdb','') \
            for x in basenames \
            if re.search(r'^\d+\.mtdb$', x)]
    master = '19991231' # arbitrary master for sorting
    for date in dates:
        dtDate = datetime.datetime.strptime(date, "%Y%m%d")
        dtMast = datetime.datetime.strptime(master, "%Y%m%d")
        if dtDate > dtMast:
            master = date
    if master == '19991231': # if it is the arbitrary start
        eprint('\nERROR: master db not found in ' + path \
              + '. Have you initialized MycoDB?', flush = True)
        return None
    master_path = formatPath( '$' + path + '/' + master + '.mtdb')

    return master_path
 
# imports database, converts into df
# returns database dataframe
def db2df(data, stdin = False):
    import pandas as pd, pandas
    columns = mtdb.columns
    if not stdin:
        data = formatPath( data )
        db_df = pd.read_csv( data, sep='\t' )
        if 'ome' not in set( db_df.columns ) and 'assembly_acc' not in set( db_df.columns ):
            db_df = pd.read_csv( data, sep = '\t', header = None )
    else:
        db_df = pd.read_csv( StringIO(data), sep='\t' )
        if 'ome' not in set( db_df.columns ) and 'assembly_acc' not in set( db_df.columns ):
            db_df = pd.read_csv( StringIO(data), sep = '\t', header = None )

    if len(db_df.keys()) == 16: # legacy conversion TO BE DEPRECATED
        eprint('\tWARNING: Legacy MycotoolsDB format will be removed in the future.', flush = True)
        db_df.columns = [
            'ome', 'genus', 'species', 'strain', 'version',
            'biosample', 'fna', 'faa', 'gff3', 'taxonomy', 'ecology',
            'eco_conf',
            'source', 'published', 'assembly_acc', 'acquisition_date'
            ]
    else:
        db_df.columns = columns
    for i, row in db_df.iterrows():
        db_df.at[i, 'taxonomy'] = read_tax(
            row['taxonomy']
            )
        db_df.at[i, 'taxonomy']['genus'] = row['genus']
        db_df.at[i, 'taxonomy']['species'] = \
            row['genus'] + ' ' + row['species']
    if len(db_df) == 16: # LEGACY conversion to be deprecated
        del db_df['ecology']
        del db_df['eco_conf']
    for i, row in db_df.iterrows(): 
        # if malformatted due to decreased entries in some lines, this will raise an IndexError
        if not row['fna'] or row['fna'] == row['ome'] + '.fa': # abbreviated line w/o file coordinates
           db_df.at[i, 'fna'] = os.environ['MYCOFNA'] + '/' + row['ome'] + '.fa'
           db_df.at[i, 'faa'] = os.environ['MYCOFAA'] + '/' + row['ome'] + '.aa.fa'
           db_df.at[i, 'gff3'] = os.environ['MYCOGFF3'] + '/' + row['ome'] + '.gff3'
        else: # has file coordinates
           db_df.at[i, 'fna'] = formatPath(row['fna'])
           db_df.at[i, 'faa'] = formatPath(row['faa'])
           db_df.at[i, 'gff3'] = formatPath(row['gff3'])

    return db_df


def df2std( df ):
    '''
    Standardized organization of database columns. DEPRECATED pandas MTDB
    '''
    # temporary check, will remove when all scripts account for new biosample column
    trans_df = df[
        mtdb.columns
        ]
    return trans_df


# imports dataframe and organizes it in accord with the `std df` format
# exports as database file into `db_path`. If rescue is set to 1, save in home folder if output doesn't exist
# if rescue is set to 0, do not output database if output dir does not exit
def df2db(df, db_path, header = False, overwrite = False, std_col = True, rescue = True):
    import pandas as pd, pandas

    df = df.set_index('ome')
    df = df.sort_index()
    df = df.reset_index()

    if db_path != sys.stdout:
        db_path = formatPath( db_path )
    elif overwrite:
        number = 0
        while os.path.exists(db_path):
            number += 1
            db_path = os.path.normpath(db_path) + '_' + str(number)

    if std_col:
        df = df2std( df )

    while True:
        try:
            df.to_csv(db_path, sep = '\t', index = None, header = header)
            break
        except FileNotFoundError:
            if rescue:
                eprint('\nOutput directory does not exist. Attempting to save in home folder.', flush = True)
                db_path = '~/' + os.path.basename(os.path.normpath(db_path))
                df.to_csv(db_path, sep = '\t', index = None)
                raise FileNotFoundError
                break
            else:
                eprint('\nOutput directory does not exist. Rescue not enabled.', flush = True)
                raise FileNotFoundError
                break


def hit2taxonomy( 
    taxid, 
    rank = 'kingdom', 
    lineage = 'fungi', 
    skip = False,
    email = None,
    api = None
    ):
    ''' 
    Takes a searchTerm string, queries NCBI via Entrez, obtains TaxIDs,
    if there are multiple hits return an error an query the TaxID for 
    taxonomy to find a fungal hit. If there is a fungal hit, create a
    tax_dict and return it.
    tax_dict = { taxonomicRank: rankName }
    '''

    tax_dict, count, sleep = {}, 0, False
    while True:
        try:
            tax_handle = Entrez.efetch( db = "Taxonomy", id = taxid, retmode = "xml" )
            break
        except HTTPError as goon:
            count += 1
            if count == 5 and skip:
                print( '\n5 failed HTTP queries. Is NCBI down?' , flush = True)
                sys.exit( 100 )
            if 500 <= goon.code <= 599:
                time.sleep( 1 )
            sleep = True

    count = 0
    while True:
        try:
            records = Entrez.read( tax_handle, validate = False )
            break
        except Entrez.Parser.NotXMLError:
            time.sleep(1)
            count += 1
            if count == 5:
                eprint( '\nERROR: 5 failed taxonomy queries. ' + str(taxid) , flush = True)
                eprint( tax_handle , flush = True)
                for line in tax_handle:
                    print( line , flush = True)
                if skip:
                    sys.exit( 1 )
                records = False
                break
            tax_handle = Entrez.efetch( db = "Taxonomy", id = taxid, retmode = "xml" )
            if 'latin-1' in str(tax_handle):
                count = 0
            
                eprint( '\tERROR: latin-1 encoding' , flush = True)
                for line in tax_handle:
                    print( line , flush = True)
                time.sleep( 300 )
                Entrez.email = email
                Entrez.api_key = api
 #               wrapper = io.TextIOWrapper(
  #                  io.BytesIO(),
   #                 encoding = 'utf-8',
    #                line_buffering = True,
     #               )
      #          for line in tax_handle:
       #             wrapper.write(line + '\n')
        #        tax_handle = wrapper
            sleep = True

    if records:
        if 'LineageEx' in records[0]:
            lineages = records[0][ 'LineageEx' ]
            for tax in lineages:
                if tax['Rank'].lower() == rank.lower():
                    tax_dict = { str(x['Rank']): str(x['ScientificName']) for x in lineages }
                    continue

    return tax_dict, sleep


# gather taxonomy by querying NCBI
# if `api_key` is set to `1`, it assumes the `Entrez.api` method has been called already
def gather_taxonomy(df, api_key = None, king='fungi', ome_index = 'ome'):

    print('\nAssimilating taxonomy', flush = True)
    if isinstance(df, mtdb):
        df = df.set_index('ome')
        tax_dicts = {v['genus']: read_tax(v['taxonomy']) for k, v in df.items()}
    else:
        df['taxonomy'] = df['taxonomy'].replace( np.nan, None )
        tax_dicts = {x['genus']: read_tax(x['taxonomy']) for i,x in df.iterrows()}

    count = 0

    for genus in tax_dicts:
        if any(tax_dicts[genus][x] for x in tax_dicts[genus]):
            continue
        print('\t' + genus, flush = True)
# if there is no api key, sleep for a second after 3 queries
# if there is an api key, sleep for a second after 10 queries
        if not api_key and count == 2:
            time.sleep(1)
            count = 0
        elif api_key and count == 9:
            time.sleep(1)
            count = 0

# gather taxonomy from information present in the genus column and query NCBI using Entrez
        ids = None
        search_term = str(genus) + ' [GENUS]'
        handle = Entrez.esearch(db='Taxonomy', term=search_term)
        ids = Entrez.read(handle)['IdList']
        count += 1
        if len(ids) == 0:
            print('\t\tNo taxonomy information', flush = True)
            continue

# for each taxID acquired, fetch the actual taxonomy information
        taxid = 0
        for tax in ids:
            if not api_key and count == 2:
                time.sleep(1)
                count = 0
            elif api_key and count == 9:
                time.sleep(1)
                count = 0
        count += 1
        for attempt in range(3):
            try:
                handle = Entrez.efetch(db="Taxonomy", id=tax, remode="xml")
                records = Entrez.read(handle)
                lineages = records[0]['LineageEx']
                break
            except IndexError:
                lineages = []

# for each lineage, if it is a part of the kingdom, `king`, use that TaxID
# if there are multiple TaxIDs, use the first one found
        if king:
            for lineage in lineages:
                if lineage['Rank'] == 'kingdom':
                    if lineage['ScientificName'].lower() == king:
                        taxid = tax
                        if len(ids) > 1:
                            print('\t\tTax ID(s): ' + str(ids), flush = True)
                        break
        else:
            for lineage in lineages:
                taxid = tax 
       
        if taxid == 0:
            print('\t\tNo taxonomy information', flush = True)
            continue

# for each taxonomic classification, add it to the taxonomy dictionary string
# append each taxonomy dictionary string to a list of dicts
        tax_dicts[genus] = {}
        for tax in lineages:
             tax_dicts[genus][tax['Rank'].lower()] = tax['ScientificName']
    
        count += 1
        if count == 3 or count == 10:
            if not api_key:
                time.sleep( 1 )
                count = 0
            elif api_key:
                if count == 10:
                    time.sleep( 1 )
                    count = 0

    return tax_dicts


# read taxonomy by conterting the string into a dictionary using `json.loads`
def read_tax(taxonomy_string):
   
    tax_strs = ['kingdom', 'phylum', 'subphylum', 'class', 'order', 'family', 'subfamily']
    if taxonomy_string: 
        if isinstance(taxonomy_string, str):
            dict_string = taxonomy_string.replace("'",'"')
            tax_dict = json.loads(dict_string)
        else:
            tax_dict = taxonomy_string
        tax_dict = {**tax_dict, **{x: '' for x in tax_strs if x not in tax_dict}}
        return tax_dict
    else:
        return {}


# assimilate taxonomy dictionary strings and append the resulting taxonomy string dicts to an inputted database
# forbid a list of taxonomic classifications you are not interested in and return a new database
def assimilate_tax(db, tax_dicts, ome_index = 'ome', forbid={'no rank', 'superkingdom', 'subkingdom', 'genus', 'species', 'species group', 'varietas', 'forma'}):

    genera = set(db['genus'])
    tax_dicts = {x: tax_dicts[x] for x in tax_dicts if tax_dicts[x]}
    for genus in tax_dicts:
        tax_dicts[genus] = {x: tax_dicts[genus][x] for x in tax_dicts[genus] if x not in forbid}
    missing = list(genera.difference(set(tax_dicts.keys())))
    
    for miss in missing:
        tax_dicts[miss] = {}
    if isinstance(db, mtdb):
        for i, genus in enumerate(db['genus']):
            db['taxonomy'][i] = tax_dicts[genus]
        return mtdb(db)
    else:
        db['taxonomy'] = db.apply(
            lambda x: tax_dicts[x['genus']], axis = 1
            )

    return db


# extract taxonomy and return a database with just the taxonomy you are interested in
def extract_tax(db, taxonomy, classification, inverse = False ):
#    import pandas as pd, pandas

    if isinstance(taxonomy, str):
        taxonomy = [taxonomy]

    taxonomy = set(x.lower() for x in taxonomy)
#    new_db = pd.DataFrame()
    db = db.set_index()

# `genus` is a special case because it has its own column, so if it is not a genus, we need to read the taxonomy
# once it is read, we can parse the taxonomy dictionary via the inputted `classification` (taxonomic classification)
    if classification != 'genus':
        new_db = mtdb().set_index()

        for ome in db:
            db[ome]['taxonomy'] = read_tax(db[ome]['taxonomy'])
            if classification in db[ome]['taxonomy']:
                if db[ome]['taxonomy'][classification].lower() in taxonomy and not inverse:
                    new_db[ome] = db[ome]
                elif inverse and db[ome]['taxonomy'][classification].lower() not in taxonomy:
                    new_db[ome] = db[ome]
            elif inverse:
               new_db[ome] = db[ome]

# if `classification` is `genus`, simply check if that fits the criteria in `taxonomy`
    elif classification == 'genus':
        taxonomy = set(x[0].upper() + x[1:].lower() for x in taxonomy)
        if not inverse:
            new_db = mtdb({ome: row for ome, row in db.items() if row['genus'] in taxonomy}, index =
                'ome')
        else:
            new_db = mtdb({ome: row for ome, row in db.items() if row['genus'] not in taxonomy}, index =
                'ome')

    return new_db.reset_index()


# extracts all omes from an `ome_list` or the inverse and returns the extracted database
def extract_omes(db, ome_list, index = 'ome', inverse = False):
    import pandas as pd, pandas

    new_db = pd.DataFrame()
    ref_set = set()
    ome_set = set(ome_list)
    if not inverse:
        for ome in ome_set:
            ref_set.add(ome.lower())
    else:
        for i, row in db.iterrows():
            if row[index] not in ome_set:
                ref_set.add( row[index].lower() )

    db = db.set_index(index)
    for i,row in db.iterrows():
        if row[index].lower() in ref_set:
            new_db = new_db.append(row)

    return new_db

if not masterDB():
    eprint('WARNING: MycotoolsDB not initialized', flush = True)



#! /usr/bin/env python3

import os
import sys
import random
import argparse
from mycotools.lib.kontools import file2list, intro, outro, formatPath, eprint
from mycotools.lib.dbtools import mtdb, extract_tax, extract_omes, masterDB

def main( 
    db, rank = False, unique_species = 0, lineage_list = False,
    inverse = False, lineage = False, ome_list = False,
    source = False, nonpublished = False
    ):

    if unique_species > 0:
        db = db.set_index('ome')
        keys = random.shuffle(list(db.keys()))
        prep_db0 = {x: db[x] for x in keys}
        prep_db1 = mtdb().set_index('ome')
        found = {}
        for ome, row in prep_db0.items():
            name = row['genus'].lower() + '_' + row['species'].lower()
            if name not in found:
                name[found] = 0
            name[found] += 1
            if name[found] <= unique_species:
                prep_db1[ome] = row
        db = prep_db1.reset_index()


    # if a taxonomy list is specified then open it, store each entry in a list
    # and extract each taxonomic entry based on the classification specified
    if lineage_list:
        taxonomy_list = file2list( formatPath(lineage_list) )
        if not taxonomy_list:
            sys.exit( 15 )

        if rank:
            new_db = mtdb()
            new_db = new_db.append( 
                extract_tax( 
                    db, taxonomy_list, 
                    rank, inverse = inverse 
                    ) 
                )

    # if there is one taxonomic group specified, extract it 
    elif lineage:
        new_db = extract_tax(
            db, lineage, rank, inverse = inverse )

    # if an ome list is specified then open it, store each entry in a list and pull each ome
    elif ome_list:
        omes = set(file2list( formatPath(ome_list) ))
        db = db.set_index('ome')
        new_db = mtdb()
        new_db = new_db.set_index()
        for i in db:
            if i in omes:
                new_db[i] = db[i]
        new_db = new_db.reset_index() 
#        new_db = extract_omes( db, omes, inverse = inverse )

    # if none of these are specified then create a `new_db` variable to work for later
    else:
        new_db = db

    # if there is a source specified, extract it or the opposite if inverse is specified
    if source and not inverse:
        new_db = new_db[new_db['source'] == source]
    elif source and inverse:
        new_db = new_db[new_db['source'] != source]

    # if you want publisheds, then just pull those out 
    new_db = new_db.set_index()
    if not nonpublished:
        todel = []
        for ome in new_db:
            if not new_db[ome]['published']:
#                new_db.at[i, 'published'] = 0
#                new_db = new_db.drop(i)
                todel.append(ome)
 #           else:
#                new_db['published'][i] = .at[i, 'published'] = row['published']
        for v in todel:
            del new_db[v]

    return new_db


if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = \
       'Extracts a MycotoolsDB from arguments. E.g.\t`extractDB.py ' + \
       '-l Atheliaceae -r family`' )
    parser.add_argument( '-d', '--database', default = masterDB(), help = 'DEFAULT: masterdb' )
    parser.add_argument( '-l', '--lineage' )
    parser.add_argument( '-r', '--rank', help = "Taxonomy rank" )
    parser.add_argument( '-s', '--source', help = 'Data source' )
    parser.add_argument( '-n', '--nonpublished', action = 'store_true', help = 'Include ' + \
        'restricted-use' )
    parser.add_argument( '-u', '--unique', default = 0, type = int, help = 'Number of same species allowed' )
    parser.add_argument( '-i', '--inverse', action = 'store_true', help = 'Inverse arguments' )
    parser.add_argument( '--headers', action = 'store_true', help = 'Include header')
    parser.add_argument( '-o', '--output' )
    parser.add_argument( '-', '--stdin', action = 'store_true', help = "Pipe MycotoolsDB from stdin" )
    parser.add_argument( '-ol', '--ome', help = "File w/list of omes" )
    parser.add_argument( '-ll', '--lineages', help = 'File w/list of lineages (same rank)' )
    args = parser.parse_args()
    db_path = formatPath( args.database )


    if (args.lineage or args.lineages) and not args.rank:
        eprint('\nERROR: need rank for lineage(s)', flush = True)
        sys.exit(5)
    elif args.lineage or args.lineages:
        eprint("\nWARNING: extracting taxonomy is subject to taxonomic error and errors in NCBI's hiearchy\n")

    output = 'stdout'
    if args.output:
        output = formatPath( args.output )
        if not output.endswith( '/' ):
            tag = ''
                       
            if args.lineage:
                tag += '_' + args.lineage
            if args.lineages:
                tag += '_taxonomy'
            if args.source:
                tag += args.source.lower()
            if not args.nonpublished:
                tag += '_pub'
            output += '/' + os.path.basename( db_path ) + tag
        
    args_dict = {
        'database': db_path,
        'output': output,
        'lineage': args.lineage,
        'lineages list': args.lineages,
        'rank': args.rank,
        'ome_list': args.ome,
        'Source': args.source,
        'unique': args.unique,
        'nonpublished': args.nonpublished,
        'inverse': args.inverse,
        'headers': bool(args.headers)
    	}
#    start_time = intro( 'Abstract Database', args_dict, stdout = args.output )

    if args.stdin:
        data = ''
        for line in sys.stdin:
            data += line.rstrip() + '\n'
        data = data.rstrip()
        db = mtdb(data, stdin=True)
    else:
       db = mtdb( db_path )
    new_db = main( 
        db, rank = args.rank, lineage = args.lineage, lineage_list = args.lineages,
        ome_list = args.ome, source = args.source, unique_species = args.unique, 
        nonpublished = args.nonpublished, inverse = args.inverse
        )
    if args.output:
        new_db.df2db( output )
    else:
        new_db.df2db( headers = bool(args.headers) )

 #   outro( start_time, stdout = args.output )
    sys.exit(0)
